
# import pytest
# import pandas as pd

# import rcv_cruncher.cvr as cvr

# params = [
#     (
#         {
#             'ranks': [['A', 'B', cvr.BallotMarks.SKIPPEDRANK]],
#             'weight': [1]
#         },
#         pd.DataFrame({
#             'rank_limit': [3],
#             'first_round_overvote': [0],
#             'ranked_single': [0],
#             'ranked_multiple': [1],
#             'ranked_3_or_more': [0],
#             'total_fully_ranked': [1],
#             'includes_duplicate_ranking': [0],
#             'includes_skipped_ranking': [0],
#             'includes_overvote_ranking': [0],
#             'total_ballots': [1],
#             'total_irregular': [0],
#             'total_undervote': [0],
#             'mean_rankings_used': [2],
#             'median_rankings_used': [2]
#         })
#     ),
#     (
#         {
#             'ranks': [
#                 ['A', 'B', cvr.BallotMarks.SKIPPEDRANK],
#                 [cvr.BallotMarks.SKIPPEDRANK, cvr.BallotMarks.OVERVOTE, 'C']
#                 ],
#             'weight': [1, 1]
#         },
#         pd.DataFrame({
#             'rank_limit': [3],
#             'first_round_overvote': [1],
#             'ranked_single': [1],
#             'ranked_multiple': [1],
#             'ranked_3_or_more': [0],
#             'total_fully_ranked': [0],
#             'includes_duplicate_ranking': [0],
#             'includes_skipped_ranking': [1],
#             'includes_overvote_ranking': [1],
#             'total_ballots': [2],
#             'total_irregular': [1],
#             'total_undervote': [0],
#             'mean_rankings_used': [1.5],
#             'median_rankings_used': [1.5]
#         })
#     ),
#     (
#         {
#             'ranks': [
#                 ['A', 'B', 'B'],
#                 ['A', 'B', 'C'],
#                 [cvr.BallotMarks.SKIPPEDRANK, cvr.BallotMarks.SKIPPEDRANK, cvr.BallotMarks.SKIPPEDRANK]
#                 ],
#             'weight': [1, 1, 1]
#         },
#         pd.DataFrame({
#             'rank_limit': [3],
#             'first_round_overvote': [0],
#             'ranked_single': [0],
#             'ranked_multiple': [2],
#             'ranked_3_or_more': [2],
#             'total_fully_ranked': [1],
#             'includes_duplicate_ranking': [1],
#             'includes_skipped_ranking': [0],
#             'includes_overvote_ranking': [0],
#             'total_ballots': [3],
#             'total_irregular': [1],
#             'total_undervote': [1],
#             'mean_rankings_used': [3],
#             'median_rankings_used': [3]
#         })
#     )
# ]


# @pytest.mark.parametrize("cvr_input, stats_expected", params)
# def test_cvr(cvr_input, stats_expected):

#     cast_vote_record = cvr.CastVoteRecord(parsed_cvr=cvr_input,
#                                           jurisdiction="x",
#                                           state="x",
#                                           year="x",
#                                           date="x",
#                                           office="x")

#     computed_stats = cast_vote_record.cvr_stats()

#     for stat in stats_expected.columns:
#         assert stats_expected[stat].item() == computed_stats[stat].item()



from __future__ import annotations
from typing import (Dict, Iterable, Optional)

import copy


class BallotMarks:

    SKIPPED = 'skipped'
    OVERVOTE = 'overvote'
    WRITEIN = 'writein'

    writein_partial_match_words = ['write']
    writein_anycase_exact_match_words = ['uwi']

    UNDERVOTE = 'undervote'
    PRETALLY_EXHAUST = 'pretally_exhaust'
    MAYBE_EXHAUSTED = 'maybe_exhausted'
    MAYBE_EXHAUSTED_BY_OVERVOTE = 'maybe_exhausted_by_overvote'
    MAYBE_EXHAUSTED_BY_REPEATED_SKIPPED_RANKING = 'maybe_exhausted_by_repeated_skipped_ranking'
    MAYBE_EXHAUSTED_BY_DUPLICATE_RANKING = 'maybe_exhausted_by_duplicate_ranking'

    @staticmethod
    def new_rule_set(combine_writein_marks: bool = False,
                     exclude_writein_marks: bool = False,
                     exclude_duplicate_candidate_marks: bool = False,
                     exclude_overvote_marks: bool = False,
                     exclude_skipped_marks: bool = False,
                     treat_combined_writeins_as_exhaustable_duplicates: bool = False,
                     exhaust_on_duplicate_candidate_marks: bool = False,
                     exhaust_on_overvote_marks: bool = False,
                     exhaust_on_repeated_skipped_marks: bool = False) -> Dict:
        return {
            'combine_writein_marks': combine_writein_marks,
            'exclude_writein_marks': exclude_writein_marks,
            'exclude_duplicate_candidate_marks': exclude_duplicate_candidate_marks,
            'exclude_overvote_marks': exclude_overvote_marks,
            'exclude_skipped_marks': exclude_skipped_marks,
            'treat_combined_writeins_as_exhaustable_duplicates': treat_combined_writeins_as_exhaustable_duplicates,
            'exhaust_on_duplicate_candidate_marks': exhaust_on_duplicate_candidate_marks,
            'exhaust_on_overvote_marks': exhaust_on_overvote_marks,
            'exhaust_on_repeated_skipped_marks': exhaust_on_repeated_skipped_marks
        }

    @staticmethod
    def check_writein_match(mark):
        matched = False

        if not matched:
            for writein_mark in BallotMarks.writein_partial_match_words:
                if writein_mark in mark.lower():
                    matched = True
                    break

        if not matched:
            for writein_mark in BallotMarks.writein_anycase_exact_match_words:
                if writein_mark == mark.lower():
                    matched = True
                    break

        return matched

    @staticmethod
    def combine_writein_marks(ballot_marks: BallotMarks) -> BallotMarks:

        if not isinstance(ballot_marks, BallotMarks):
            raise TypeError('ballot_marks must be BallotMarks object.')

        copy_ballot_marks = copy.deepcopy(ballot_marks)
        new_marks = [BallotMarks.WRITEIN if BallotMarks.check_writein_match(mark) else mark
                     for mark in copy_ballot_marks.marks]
        copy_ballot_marks.update_marks(new_marks)
        return copy_ballot_marks

    @staticmethod
    def remove_mark(ballot_marks: BallotMarks, remove_marks: Iterable) -> BallotMarks:

        if not isinstance(ballot_marks, BallotMarks):
            raise TypeError('ballot_marks must be BallotMarks object.')

        if isinstance(remove_marks, str):
            raise TypeError('remove_marks must be Iterable, but cannot be string.')

        copy_ballot_marks = copy.deepcopy(ballot_marks)
        new_marks = copy_ballot_marks.marks
        for remove_mark in remove_marks:
            new_marks = [mark for mark in new_marks if mark != remove_mark]
        copy_ballot_marks.update_marks(new_marks)
        return copy_ballot_marks

    @staticmethod
    def remove_duplicate_candidate_marks(ballot_marks: BallotMarks) -> BallotMarks:

        if not isinstance(ballot_marks, BallotMarks):
            raise TypeError('ballot_marks must be BallotMarks object.')

        copy_ballot_marks = copy.deepcopy(ballot_marks)
        new_marks_list = []
        new_marks_set = set()
        for mark in copy_ballot_marks.marks:
            if mark not in new_marks_set.difference({BallotMarks.OVERVOTE, BallotMarks.SKIPPED}):
                new_marks_list.append(mark)
                new_marks_set = new_marks_set.union({mark})
        copy_ballot_marks.update_marks(new_marks_list)
        return copy_ballot_marks

    def __init__(self, marks: Optional[Iterable] = None) -> None:

        self.marks = []
        self.unique_marks = {}
        self.unique_candidates = {}

        if marks:
            self.update_marks(marks)

        self.rules = {}
        self.inactive_type = None

        # in the future, add functionality to handle overvotes that can be conditionally resolved
        # if all but one of the overvoted candidates is eliminated before the overvote is reached

    def update_marks(self, new_marks: Iterable) -> None:

        if isinstance(new_marks, str):
            raise TypeError('new_marks must be Iterable, but cannot be string.')

        self.marks = list(new_marks)
        self.unique_marks = set(self.marks)
        self.unique_candidates = self.unique_marks - {BallotMarks.SKIPPED, BallotMarks.OVERVOTE}

    def apply_rules(self,
                    combine_writein_marks: bool = False,
                    exclude_writein_marks: bool = False,
                    exclude_duplicate_candidate_marks: bool = False,
                    exclude_overvote_marks: bool = False,
                    exclude_skipped_marks: bool = False,
                    treat_combined_writeins_as_exhaustable_duplicates: bool = False,
                    exhaust_on_duplicate_candidate_marks: bool = False,
                    exhaust_on_overvote_marks: bool = False,
                    exhaust_on_repeated_skipped_marks: bool = False) -> None:

        if self.rules:
            raise RuntimeError('rules have already been applied to this ballot')

        self.rules = {
            'combine_writein_marks': combine_writein_marks,
            'exclude_writein_marks': exclude_writein_marks,
            'exclude_duplicate_candidate_marks': exclude_duplicate_candidate_marks,
            'exclude_overvote_marks': exclude_overvote_marks,
            'exclude_skipped_marks': exclude_skipped_marks,
            'treat_combined_writeins_as_exhaustable_duplicates': treat_combined_writeins_as_exhaustable_duplicates,
            'exhaust_on_duplicate_candidate_marks': exhaust_on_duplicate_candidate_marks,
            'exhaust_on_overvote_marks': exhaust_on_overvote_marks,
            'exhaust_on_repeated_skipped_marks': exhaust_on_repeated_skipped_marks
        }

        all_skipped = None
        specific_exhaust = None

        if self.unique_marks == {BallotMarks.SKIPPED}:
            all_skipped = True
        else:
            all_skipped = False

        # has to occur before exhaustion by duplicates is computed
        if combine_writein_marks and treat_combined_writeins_as_exhaustable_duplicates:
            self.update_marks(self.combine_writein_marks(self).marks)

        new_marks_list = []
        new_marks_set = set()
        for mark_idx, mark in enumerate(self.marks):

            repeated_skipped_marks_present = False
            if mark_idx + 1 < len(self.marks):

                remaining_marks = self.marks[mark_idx+1:]
                next_mark = remaining_marks[0]

                # check for successive skips
                if mark == BallotMarks.SKIPPED and next_mark == BallotMarks.SKIPPED:
                    repeated_skipped_marks_present = True
                else:
                    repeated_skipped_marks_present = False

                # and check if any non-skip marks remain after skips
                if repeated_skipped_marks_present and set(remaining_marks) and set(remaining_marks) != {BallotMarks.SKIPPED}:
                    repeated_skipped_marks_present = True
                else:
                    repeated_skipped_marks_present = False

            if exhaust_on_repeated_skipped_marks and repeated_skipped_marks_present:
                if not self.inactive_type:
                    specific_exhaust = self.MAYBE_EXHAUSTED_BY_REPEATED_SKIPPED_RANKING
                break

            overvote_mark_present = mark == BallotMarks.OVERVOTE
            if exhaust_on_overvote_marks and overvote_mark_present:
                if not self.inactive_type:
                    specific_exhaust = self.MAYBE_EXHAUSTED_BY_OVERVOTE
                break

            duplicate_candidate_mark = mark in new_marks_set.difference({BallotMarks.OVERVOTE, BallotMarks.SKIPPED})
            if exhaust_on_duplicate_candidate_marks and duplicate_candidate_mark:
                if not self.inactive_type:
                    specific_exhaust = self.MAYBE_EXHAUSTED_BY_DUPLICATE_RANKING
                break

            new_marks_list.append(mark)
            new_marks_set = new_marks_set.union({mark})

        self.update_marks(new_marks_list)

        if combine_writein_marks and not treat_combined_writeins_as_exhaustable_duplicates:
            self.update_marks(self.combine_writein_marks(self).marks)

        if exclude_duplicate_candidate_marks:
            self.update_marks(self.remove_duplicate_candidate_marks(self).marks)

        if exclude_overvote_marks:
            self.update_marks(self.remove_mark(self, [BallotMarks.OVERVOTE]).marks)

        if exclude_skipped_marks:
            self.update_marks(self.remove_mark(self, [BallotMarks.SKIPPED]).marks)

        if exclude_writein_marks:
            self.update_marks(self.remove_mark(self, [BallotMarks.WRITEIN]).marks)

        if all_skipped:
            self.inactive_type = self.UNDERVOTE
        elif not self.marks:
            self.inactive_type = self.PRETALLY_EXHAUST
        elif specific_exhaust:
            self.inactive_type = specific_exhaust
        else:
            self.inactive_type = self.MAYBE_EXHAUSTED




import copy

import pandas as pd

import rcv_cruncher.util as util

from rcv_cruncher.marks import BallotMarks


class CastVoteRecord_tables:

    # EXPORT

    def get_cvr_table(self, table_format: str = "rank") -> pd.DataFrame:

        if table_format != "rank" and table_format != "candidate":
            raise RuntimeError('table_format argument must be "rank" or "candidate"')

        if table_format == "rank":
            return self._rank_header_cvr()
        elif table_format == "candidate":
            return self._candidate_header_cvr()

    def _rank_header_cvr(self) -> pd.DataFrame:

        ballot_dict = copy.deepcopy(self.get_cvr_dict())
        bs = [b.marks for b in ballot_dict['ballot_marks']]
        weight = ballot_dict['weight']
        del ballot_dict['ballot_marks']
        del ballot_dict['weight']

        # how many ranks?
        num_ranks = max(len(i) for i in bs)

        # make sure all ballots are lists of equal length, adding trailing 'skipped' if necessary
        bs = [b + ([BallotMarks.SKIPPED] * (num_ranks - len(b))) for b in bs]

        # assemble output_table, start with extras
        output_df = pd.DataFrame.from_dict(ballot_dict)

        # are weights all one, then dont add to output
        if not all([i == 1 for i in weight]):
            output_df['weight'] = [float(w) for w in weight]

        # add in rank columns
        for i in range(1, num_ranks + 1):
            output_df['rank' + str(i)] = [b[i-1] for b in bs]

        return output_df

    def _candidate_header_cvr(self) -> pd.DataFrame:

        # get ballots and candidates
        ballot_dl = copy.deepcopy(self.get_cvr_dict())
        candidates = copy.deepcopy(self.get_candidates()).unique_candidates
        candidates.update({BallotMarks.OVERVOTE})

        # remove weights if all equal to 1
        if set(ballot_dl['weight']) == {1}:
            del ballot_dl['weight']

        # add rank limit
        ballot_dl['rank_limit'] = [len(ballot_dl['ballot_marks'][0].marks)] * len(ballot_dl['ballot_marks'])

        # convert dict of list to list of dicts
        ballot_ld = util.DL2LD(ballot_dl)

        # add candidate index information
        for b in ballot_ld:
            b.update({f"candidate_{cand}": None for cand in candidates})
            for rank_idx, cand in enumerate(b['ballot_marks'].marks, start=1):
                if cand != BallotMarks.SKIPPED:
                    if not b[f"candidate_{cand}"]:
                        b[f"candidate_{cand}"] = str(rank_idx)
                    else:
                        b[f"candidate_{cand}"] += f',{rank_idx}'
            del b['ballot_marks']

        df = pd.DataFrame(ballot_ld)
        return df.reindex(sorted(df.columns), axis=1)


# def cumulative_ranking_tables(ctx):
#     """
#     Return cumulative ranking tables. Rows are candidate names and columns are rank numbers.
#     Reading across columns, the tables show the accumulating count/percentage of ballots that marked
#     a candidate as more ranks are considered. The final column shows the count/percentage of ballots
#     that never marked the candidate.
#     """

#     # get inputs
#     candidate_set = sorted(ballots.candidates(ctx))

#     # ballot rank limit
#     ballot_length = len(ballots.input_ballots(ctx)['ranks'][0])

#     # get cleaned ballots
#     cleaned_dict = copy.deepcopy(ballots.cleaned_ballots(ctx))
#     ballot_set = [{'ranks': ranks + (['NA'] * (ballot_length - len(ranks))), 'weight': weight}
#                   for ranks, weight in zip(cleaned_dict['ranks'], cleaned_dict['weight'])]

#     # total ballots
#     total_ballots = sum([d['weight'] for d in ballot_set])

#     # create data frame that will be populated and output
#     col_names = ["Rank " + str(i + 1) for i in range(ballot_length)] + ['Did Not Rank']
#     cumulative_percent_df = pd.DataFrame(util.NAN, index=candidate_set, columns=col_names)
#     cumulative_count_df = pd.DataFrame(util.NAN, index=candidate_set, columns=col_names)

#     # tally candidate counts by rank
#     rank_counts = []
#     for rank in range(0, ballot_length):
#         rank_cand_set = set([b['ranks'][rank] for b in ballot_set]) - {'NA'}
#         current_rank_count = {cand: 0 for cand in rank_cand_set}
#         for cand in rank_cand_set:
#             current_rank_count[cand] = sum(b['weight'] for b in ballot_set if cand == b['ranks'][rank])
#         rank_counts.append(current_rank_count)

#     # accumulate ballot counts that rank candidates
#     cumulative_counter = {cand: 0 for cand in candidate_set}
#     for rank in range(0, ballot_length):
#         for cand in candidate_set:
#             # if candidate has any marks for this rank, accumulate them
#             if cand in rank_counts[rank]:
#                 cumulative_counter[cand] += rank_counts[rank][cand]
#             # update tables
#             cumulative_count_df.loc[cand, 'Rank ' + str(rank + 1)] = cumulative_counter[cand]
#             cumulative_percent_df.loc[cand, 'Rank ' + str(rank + 1)] = cumulative_counter[cand] * 100 / total_ballots

#     # fill in Did Not Rank column
#     for cand in candidate_set:
#         cumulative_count_df.loc[cand, 'Did Not Rank'] = total_ballots - cumulative_counter[cand]
#         cumulative_percent_df.loc[cand, 'Did Not Rank'] = (total_ballots - cumulative_counter[cand]) * 100 / total_ballots

#     cumulative_count_df = cumulative_count_df.astype(float).round(3)
#     cumulative_percent_df = cumulative_percent_df.astype(float).round(3)

#     return cumulative_count_df, cumulative_percent_df

# def first_second_tables(ctx):
#     """
#     Return two pandas tables with candidates as columns and first row showing distribution of first round votes.
#     Subsequent rows indicate second choice vote distribution for each column.

#     first table is vote counts
#     second table is percentages
#     """

#     candidate_set = sorted(ballots.candidates(ctx))
#     cleaned_dict = copy.deepcopy(ballots.cleaned_ballots(ctx))
#     ballot_set = [{'ranks': ranks, 'weight': weight}
#                   for ranks, weight in zip(cleaned_dict['ranks'], cleaned_dict['weight'])]

#     # create data frame that will be populated and output
#     percent_no_exhaust_df = pd.DataFrame(util.NAN, index=['first_choice', *candidate_set], columns=candidate_set)
#     percent_df = pd.DataFrame(util.NAN, index=['first_choice', *candidate_set, 'exhaust'], columns=candidate_set)
#     count_df = pd.DataFrame(util.NAN, index=['first_choice', *candidate_set, 'exhaust'], columns=candidate_set)

#     # group ballots by first choice
#     first_choices = {cand: [] for cand in candidate_set}
#     for b in ballot_set:
#         if len(b['ranks']) >= 1:
#             first_choices[b['ranks'][0]].append(b)

#     # sum total first round votes
#     total_first_round_votes = 0
#     for cand in first_choices:
#         total_first_round_votes += sum([b['weight'] for b in first_choices[cand]])

#     # add first choices to tables
#     # and calculate second choices
#     for cand in candidate_set:

#         ############################################################
#         # update first round table values
#         first_choice_count = sum([b['weight'] for b in first_choices[cand]])
#         first_choice_percent = (first_choice_count / total_first_round_votes) * 100

#         count_df.loc['first_choice', cand] = first_choice_count
#         percent_df.loc['first_choice', cand] = first_choice_percent
#         percent_no_exhaust_df.loc['first_choice', cand] = first_choice_percent

#         ############################################################
#         # calculate second choices, group second choices by candidate
#         possible_second_choices = list(set(candidate_set) - {cand})
#         second_choices = {backup_cand: [] for backup_cand in possible_second_choices + ['exhaust']}

#         # group ballots by second choices
#         for b in first_choices[cand]:
#             if len(b['ranks']) >= 2:
#                 second_choices[b['ranks'][1]].append(b['weight'])
#             else:
#                 second_choices['exhaust'].append(b['weight'])

#         # sum total second round votes
#         total_second_choices = 0
#         total_second_choices_no_exhaust = 0
#         for backup_cand in second_choices:
#             total_second_choices += sum(second_choices[backup_cand])
#             if backup_cand != 'exhaust':
#                 total_second_choices_no_exhaust += sum(second_choices[backup_cand])

#         # count second choices and add to table
#         for backup_cand in second_choices:

#             # fill in second choice values in table
#             second_choice_count = sum(second_choices[backup_cand])

#             # if there are not backup votes fill with zeros
#             if total_second_choices == 0:
#                 second_choice_percent = 0
#             else:
#                 second_choice_percent = (second_choice_count / total_second_choices) * 100

#             if total_second_choices_no_exhaust == 0:
#                 second_choice_percent_no_exhaust = 0
#             else:
#                 second_choice_percent_no_exhaust = (second_choice_count / total_second_choices_no_exhaust) * 100

#             count_df.loc[backup_cand, cand] = second_choice_count
#             percent_df.loc[backup_cand, cand] = second_choice_percent
#             if backup_cand != 'exhaust':
#                 percent_no_exhaust_df.loc[backup_cand, cand] = second_choice_percent_no_exhaust

#     count_df = count_df.astype(float).round(3)
#     percent_df = percent_df.astype(float).round(3)
#     percent_no_exhaust_df = percent_no_exhaust_df.astype(float).round(3)

#     return count_df, percent_df, percent_no_exhaust_df

# def rank_usage_tables(ctx):
#     """
#     DOES NOT USE BALLOT WEIGHTS
#     """
#     candidate_set = sorted(ballots.candidates(ctx, exclude_writeins=False))

#     # remove skipped ranks
#     ballot_set = [util.remove(util.BallotMarks.SKIPPEDRANK, b) for b in ballots.input_ballots(ctx)['ranks']]

#     # remove empty ballots and those that start with overvote
#     ballot_set = [b for b in ballot_set if len(b) >= 1 and b[0] != util.BallotMarks.OVERVOTE]
#     # remove other overvotes
#     ballot_set = [util.remove(util.BallotMarks.OVERVOTE, b) for b in ballot_set]
#     # remove duplicate rankings
#     ballot_set = [util.remove_dup(b) for b in ballot_set]

#     all_ballots_label = "Any candidate"
#     n_ballots_label = "Number of Ballots (excluding undervotes and ballots with first round overvote)"
#     mean_label = "Mean Valid Rankings Used (excluding duplicates)"
#     median_label = "Median Valid Rankings Used (excluding duplicates)"

#     rows = [all_ballots_label] + candidate_set
#     cols = [n_ballots_label, mean_label, median_label]
#     df = pd.DataFrame(index=rows, columns=cols)
#     df.index.name = "Ballots with first choice:"

#     n_ballots = len(ballot_set)
#     mean_rankings = sum(len(b) for b in ballot_set)/n_ballots
#     median_rankings = statistics.median(len(b) for b in ballot_set)

#     df.loc[all_ballots_label, n_ballots_label] = n_ballots
#     df.loc[all_ballots_label, mean_label] = mean_rankings
#     df.loc[all_ballots_label, median_label] = median_rankings

#     # group ballots by first choice
#     first_choices = {cand: [] for cand in candidate_set}
#     for b in ballot_set:
#         first_choices[b[0]].append(b)

#     for cand in candidate_set:
#         df.loc[cand, n_ballots_label] = len(first_choices[cand])
#         if first_choices[cand]:
#             df.loc[cand, mean_label] = sum(len(b) for b in first_choices[cand])/len(first_choices[cand])
#             df.loc[cand, median_label] = statistics.median(len(b) for b in first_choices[cand])
#         else:
#             df.loc[cand, mean_label] = 0
#             df.loc[cand, median_label] = 0

#     return df


# def crossover_table(ctx):

#     candidate_set = sorted(ballots.candidates(ctx, exclude_writeins=False))

#     ballot_dict = ballots.input_ballots(ctx)
#     ballot_weights = copy.deepcopy(ballot_dict['weight'])
#     ranks = [util.remove(util.BallotMarks.SKIPPEDRANK, b) for b in ballot_dict['ranks']]
#     ballot_set = [{'ranks': ranks, 'weight': weight}
#                   for ranks, weight in zip(ranks, ballot_weights)]

#     index_label = "Ballots with first choice:"
#     n_ballots_label = "Number of Ballots"

#     colname_dict = {cand: cand + " ranked in top 3" for cand in candidate_set}

#     rows = candidate_set
#     cols = [n_ballots_label] + list(colname_dict.values())
#     count_df = pd.DataFrame(index=rows, columns=cols)
#     count_df.index.name = index_label
#     percent_df = pd.DataFrame(index=rows, columns=cols)
#     percent_df.index.name = index_label

#     # group ballots by first choice
#     first_choices = {cand: [] for cand in candidate_set}
#     for b in ballot_set:
#         if len(b['ranks']) >= 1 and b['ranks'][0] != util.BallotMarks.OVERVOTE:
#             first_choices[b['ranks'][0]].append(b)

#     for cand in candidate_set:

#         n_first_choice = sum(b['weight'] for b in first_choices[cand])
#         count_df.loc[cand, n_ballots_label] = n_first_choice
#         percent_df.loc[cand, n_ballots_label] = n_first_choice

#         for opponent in candidate_set:

#             if n_first_choice:
#                 crossover_ballots = [True if opponent in b['ranks'][0:min(3, len(b['ranks']))] else False
#                                      for b in first_choices[cand]]
#                 crossover_val = sum(b['weight'] for b, flag in zip(first_choices[cand], crossover_ballots) if flag)
#                 count_df.loc[cand, colname_dict[opponent]] = crossover_val
#                 percent_df.loc[cand, colname_dict[opponent]] = crossover_val*100/n_first_choice
#             else:
#                 count_df.loc[cand, colname_dict[opponent]] = 0
#                 percent_df.loc[cand, colname_dict[opponent]] = 0

#     # convert decimal to float
#     count_df = count_df.astype(float).round(3)
#     percent_df = percent_df.astype(float).round(3)

#     return count_df, percent_df


# def condorcet_tables(ctx):
#     """
#     Returns a two condorcet tables as a pandas data frame with candidate names as row and column indices.
#     One contains counts and the other contains percents.

#     Each cell indicates the count/percentage of ballots that ranked the row-candidate over
#     the column-candidate (including ballots that only ranked the row-candidate). When calculating percents,
#     the denominator is each cell is the number of ballots that ranked the row-candidate OR the column-candidate.

#     Symmetric cells about the diagonal should sum to 100 (for the percent table).
#     """
#     candidate_set = sorted(ballots.candidates(ctx))
#     cleaned_dict = copy.deepcopy(ballots.cleaned_ballots(ctx))
#     ballot_set = [{'ranks': ranks, 'weight': weight}
#                   for ranks, weight in zip(cleaned_dict['ranks'], cleaned_dict['weight'])]

#     # create data frame that will be populated and output
#     condorcet_percent_df = pd.DataFrame(util.NAN, index=candidate_set, columns=candidate_set)
#     condorcet_count_df = pd.DataFrame(util.NAN, index=candidate_set, columns=candidate_set)

#     # turn ballot-lists into ballot-dict with
#     # key 'id' containing a unique integer id for the ballot
#     # key 'ranks' containing the original ballot-list
#     ballot_dicts = [{'id': ind, 'ballot': ballot} for ind, ballot in enumerate(ballot_set)]

#     # make dictionary with candidate as key, and value as list of ballot-dicts
#     # that contain their name in any rank
#     cand_ballot_dict = {cand: [ballot for ballot in ballot_dicts if cand in ballot['ballot']['ranks']]
#                         for cand in candidate_set}

#     # all candidate pairs
#     cand_pairs = itertools.combinations(candidate_set, 2)

#     for pair in cand_pairs:
#         cand1 = pair[0]
#         cand2 = pair[1]

#         # get the union of their ballots
#         combined_ballot_list = cand_ballot_dict[cand1] + cand_ballot_dict[cand2]
#         uniq_pair_ballots = list({v['id']: v['ballot'] for v in combined_ballot_list}.values())

#         uniq_pair_ballots_weights = [ballot['weight'] for ballot in uniq_pair_ballots]
#         sum_weighted_ballots = sum(uniq_pair_ballots_weights)

#         # which ballots rank cand1 above cand2?
#         cand1_vs_cand2 = [util.index_inf(b['ranks'], cand1) < util.index_inf(b['ranks'], cand2) for b in uniq_pair_ballots]
#         cand1_vs_cand2_weightsum = sum(weight * flag for weight, flag
#                                        in zip(uniq_pair_ballots_weights, cand1_vs_cand2))

#         # the remainder then must rank cand2 over cand1
#         cand2_vs_cand1 = [not i for i in cand1_vs_cand2]
#         cand2_vs_cand1_weightsum = sum(weight * flag for weight, flag
#                                        in zip(uniq_pair_ballots_weights, cand2_vs_cand1))

#         # add counts to df
#         condorcet_count_df.loc[cand1, cand2] = cand1_vs_cand2_weightsum
#         condorcet_count_df.loc[cand2, cand1] = cand2_vs_cand1_weightsum

#         # calculate percent
#         if sum_weighted_ballots:
#             cand1_percent = (cand1_vs_cand2_weightsum / sum_weighted_ballots) * 100
#             cand2_percent = (cand2_vs_cand1_weightsum / sum_weighted_ballots) * 100
#         else:
#             cand1_percent = 0
#             cand2_percent = 0

#         # add to df
#         condorcet_percent_df.loc[cand1, cand2] = cand1_percent
#         condorcet_percent_df.loc[cand2, cand1] = cand2_percent

#     # find condorcet winner and set index name to include winner
#     condorcet_winner = None

#     if len(candidate_set) == 1:

#         condorcet_winner = candidate_set[0]
#     else:

#         for cand in candidate_set:

#             not_cand = set(candidate_set) - {cand}
#             all_winner = all(condorcet_percent_df.loc[cand, not_cand] > 50)

#             if all_winner:
#                 if condorcet_winner is None:
#                     condorcet_winner = cand
#                 else:
#                     print("cannottt be more than one condorcet winner!!!!")
#                     exit(1)

#     # convert decimal to float
#     condorcet_count_df = condorcet_count_df.astype(float).round(3)
#     condorcet_percent_df = condorcet_percent_df.astype(float).round(3)

#     return condorcet_count_df, condorcet_percent_df, condorcet_winner




class RCV_tables:

    pass

    # def ballot_debug_df(self, *, tabulation_num=1):
    #     """
    #     Return pandas data frame with ranks as well stats on exhaustion, ranked_multiple ...
    #     """

    #     stat_names = [
    #         'contains_duplicate',
    #         f'pretally_exhausted{tabulation_num}',
    #         f'posttally_exhausted{tabulation_num}',
    #         f'posttally_exhausted_by_abstention{tabulation_num}',
    #         f'posttally_exhausted_by_overvote{tabulation_num}',
    #         f'posttally_exhausted_by_rank_limit{tabulation_num}',
    #         f'posttally_exhausted_by_skipped_rankings{tabulation_num}',
    #         f'posttally_exhausted_by_duplicate_rankings{tabulation_num}',
    #         'first_round_overvote',
    #         'fully_ranked',
    #         'contains_overvote',
    #         'ranked_multiple',
    #         'ranked_single',
    #         'undervote',
    #         'irregular'
    #     ]

    #     dct = {stat_name: list(self._stat_table[stat_name]) for stat_name in stat_names}
    #     dct['used_last_rank'] = self.used_last_rank()

    #     # get ballot info
    #     ballot_dict = copy.deepcopy(ballots.input_ballots(self.ctx, combine_writeins=False))
    #     bs = ballot_dict['ranks']

    #     # how many ranks?
    #     num_ranks = max(len(i) for i in bs)

    #     # make sure all ballots are lists of equal length, adding trailing 'skipped' if necessary
    #     bs = [b + ([util.BallotMarks.SKIPPEDRANK] * (num_ranks - len(b))) for b in bs]

    #     # add in rank columns
    #     ranks = {}
    #     for i in range(1, num_ranks + 1):
    #         ranks['rank' + str(i)] = [b[i - 1] for b in bs]

    #     # assemble output_table, start with extras
    #     return pd.DataFrame.from_dict({**ranks, **dct})

# def first_choice_to_finalist_table(rcv_obj):

#     dfs = []
#     for iTab in range(1, rcv_obj.n_tabulations()+1):

#         # who had any ballot weight allotted
#         finalist_candidates = list(rcv_obj.finalist_candidates(tabulation_num=iTab)) + ['exhaust']
#         candidate_set = sorted(ballots.candidates(rcv_obj.ctx))

#         ballot_set = [{'ranks': ranks, 'weight': weight, 'weight_distrib': distrib}
#                       for ranks, weight, distrib
#                       in zip(rcv_obj.get_initial_ranks(tabulation_num=iTab),
#                              rcv_obj.get_initial_weights(tabulation_num=iTab),
#                              rcv_obj.get_final_weight_distrib(tabulation_num=iTab))]

#         index_label = "Ballots with first choice:"
#         n_ballots_label = "Number of Ballots"

#         colname_dict = {cand: "% of votes to " + cand for cand in finalist_candidates}

#         rows = candidate_set
#         cols = [n_ballots_label] + list(colname_dict.values())
#         df = pd.DataFrame(index=rows, columns=cols + ['percent_sum'])
#         df.index.name = index_label

#         # group ballots by first choice
#         first_choices = {cand: [] for cand in candidate_set}
#         for b in ballot_set:
#             if len(b['ranks']) >= 1 and b['ranks'][0] in first_choices:
#                 first_choices[b['ranks'][0]].append(b)

#         for cand in candidate_set:

#             total_first_choice_ballots = sum(b['weight'] for b in first_choices[cand])
#             df.loc[cand, n_ballots_label] = total_first_choice_ballots

#             if total_first_choice_ballots:

#                 redistrib = {opponent: 0 for opponent in finalist_candidates}
#                 for b in first_choices[cand]:
#                     for el in b['weight_distrib']:
#                         if el[0] == 'empty':
#                             redistrib['exhaust'] += el[1]
#                         else:
#                             redistrib[el[0]] += el[1]

#                 redistrib_total_check = 0
#                 for opponent in redistrib:
#                     redistrib_percent = redistrib[opponent] / total_first_choice_ballots * 100
#                     df.loc[cand, colname_dict[opponent]] = redistrib_percent
#                     redistrib_total_check += redistrib_percent
#                 df.loc[cand, 'percent_sum'] = redistrib_total_check

#             else:
#                 for opponent in finalist_candidates:
#                     df.loc[cand, colname_dict[opponent]] = 0
#                 df.loc[cand, 'percent_sum'] = 0

#         df = df.astype(float).round(3)

#         dfs.append(df)

#     return dfs


# def round_by_round(rcv_obj):

#     rcv_dfs = []

#     for iTab in range(1, rcv_obj.n_tabulations()+1):

#         num_rounds = rcv_obj.n_rounds(tabulation_num=iTab)

#         first_round_exhaust = rcv_obj.total_pretally_exhausted(tabulation_num=iTab)

#         # get rcv results
#         rounds_full = [rcv_obj.get_round_tally_tuple(i, tabulation_num=iTab) for i in range(1, num_rounds + 1)]
#         transfers = [rcv_obj.get_round_transfer_dict(i, tabulation_num=iTab) for i in range(1, num_rounds + 1)]

#         # reformat contest outputs into useful dicts
#         cand_outcomes = rcv_obj.get_candidate_outcomes(tabulation_num=iTab)

#         # reorder candidate names
#         # winners in ascending order of round won
#         # followed by losers in descending order of round lost
#         reorder_dicts = []
#         for d in cand_outcomes:

#             if d['round_elected']:
#                 d['order'] = -1 * (1 / d['round_elected'])
#             else:
#                 d['order'] = 1 / d['round_eliminated']

#             reorder_dicts.append(d)

#         ordered_candidates_names = [d['name'] for d in sorted(reorder_dicts, key=lambda x: x['order'])]

#         # setup data frame
#         row_names = ordered_candidates_names + ['exhaust']
#         rcv_df = pd.DataFrame(util.NAN, index=row_names + ['colsum'], columns=['candidate'])
#         rcv_df.loc[row_names + ['colsum'], 'candidate'] = row_names + ['colsum']

#         # loop through rounds
#         for rnd in range(1, num_rounds + 1):

#             rnd_info = {rnd_cand: rnd_tally for rnd_cand, rnd_tally in zip(*rounds_full[rnd-1])}
#             rnd_info['exhaust'] = 0

#             rnd_transfer = dict(transfers[rnd-1])

#             # add round data
#             for cand in row_names:

#                 rnd_percent_col = 'r' + str(rnd) + '_active_percent'
#                 rnd_count_col = 'r' + str(rnd) + '_count'
#                 rnd_transfer_col = 'r' + str(rnd) + '_transfer'

#                 rcv_df.loc[cand, rnd_percent_col] = 100*(rnd_info[cand]/sum(rnd_info.values()))
#                 rcv_df.loc[cand, rnd_count_col] = rnd_info[cand]
#                 rcv_df.loc[cand, rnd_transfer_col] = rnd_transfer[cand]

#             # maintain cumulative exhaust total
#             if rnd == 1:
#                 rcv_df.loc['exhaust', rnd_count_col] = first_round_exhaust
#             else:
#                 last_rnd_count_col = 'r' + str(rnd-1) + '_count'
#                 last_rnd_transfer_col = 'r' + str(rnd-1) + '_transfer'
#                 current_rnd_count_val = sum(rcv_df.loc['exhaust', [last_rnd_count_col, last_rnd_transfer_col]].astype(float))
#                 rcv_df.loc['exhaust', rnd_count_col] = current_rnd_count_val

#             # sum round columns
#             rcv_df.loc['colsum', rnd_count_col] = sum(rcv_df.loc[row_names, rnd_count_col].astype(float))
#             rcv_df.loc['colsum', rnd_transfer_col] = sum(rcv_df.loc[row_names, rnd_transfer_col].astype(float))
#             rcv_df.loc['colsum', rnd_percent_col] = sum(rcv_df.loc[row_names, rnd_percent_col].astype(float))

#         # # convert from decimal to float
#         rcv_df.loc[row_names + ['colsum'], rcv_df.columns != "candidate"] = \
#             rcv_df.loc[row_names + ['colsum'], rcv_df.columns != "candidate"].astype(float).round(3)

#         # remove rownames
#         rcv_df = rcv_df.reset_index(drop=True)

#         rcv_dfs.append(rcv_df)

#     return rcv_dfs

    # def ballot_debug_df(self, *, tabulation_num=1):
    #     """
    #     Return pandas data frame with ranks as well stats on exhaustion, ranked_multiple ...
    #     """

    #     stat_names = [
    #         'contains_duplicate',
    #         f'pretally_exhausted{tabulation_num}',
    #         f'posttally_exhausted{tabulation_num}',
    #         f'posttally_exhausted_by_abstention{tabulation_num}',
    #         f'posttally_exhausted_by_overvote{tabulation_num}',
    #         f'posttally_exhausted_by_rank_limit{tabulation_num}',
    #         f'posttally_exhausted_by_skipped_rankings{tabulation_num}',
    #         f'posttally_exhausted_by_duplicate_rankings{tabulation_num}',
    #         'first_round_overvote',
    #         'fully_ranked',
    #         'contains_overvote',
    #         'ranked_multiple',
    #         'ranked_single',
    #         'undervote',
    #         'irregular'
    #     ]

    #     dct = {stat_name: list(self._stat_table[stat_name]) for stat_name in stat_names}
    #     dct['used_last_rank'] = self.used_last_rank()

    #     # get ballot info
    #     ballot_dict = copy.deepcopy(ballots.input_ballots(self.ctx, combine_writeins=False))
    #     bs = ballot_dict['ranks']

    #     # how many ranks?
    #     num_ranks = max(len(i) for i in bs)

    #     # make sure all ballots are lists of equal length, adding trailing 'skipped' if necessary
    #     bs = [b + ([util.BallotMarks.SKIPPEDRANK] * (num_ranks - len(b))) for b in bs]

    #     # add in rank columns
    #     ranks = {}
    #     for i in range(1, num_ranks + 1):
    #         ranks['rank' + str(i)] = [b[i - 1] for b in bs]

    #     # assemble output_table, start with extras
    #     return pd.DataFrame.from_dict({**ranks, **dct})



import decimal
import os
import pathlib
import csv

from rcv_cruncher.marks import BallotMarks

###############################################################
# constants

NAN = decimal.Decimal('NaN')


class InactiveType:

    UNDERVOTE = BallotMarks.UNDERVOTE
    PRETALLY_EXHAUST = BallotMarks.PRETALLY_EXHAUST

    NOT_EXHAUSTED = "not_exhausted"
    POSTTALLY_EXHAUSTED_BY_RANK_LIMIT = "posttally_exhausted_by_rank_limit"
    POSTTALLY_EXHAUSTED_BY_ABSTENTION = "posttally_exhausted_by_abstention"
    POSTTALLY_EXHAUSTED_BY_OVERVOTE = "posttally_exhausted_by_overvote"
    POSTTALLY_EXHAUSTED_BY_REPEATED_SKIPPED_RANKING = "posttally_exhausted_by_repeated_skipped_ranking"
    POSTTALLY_EXHAUSTED_BY_DUPLICATE_RANKING = "posttally_exhausted_by_duplicate_ranking"


########################
# helper funcs

class CSVLogger:

    def __init__(self, path, header_list):
        self.row_length = len(header_list)
        self.path = path
        self.file = open(path, 'w', newline='')
        self.writer = csv.writer(self.file, delimiter=',', quotechar='"', quoting=csv.QUOTE_ALL)
        self.write(header_list)

    def write(self, row_list):
        if len(row_list) != self.row_length:
            msg = f"CSVLogger.write ({self.path.name}) row list has length {len(row_list)}, "
            msg += f"doesn't match header list length ({self.row_length})"
            raise RuntimeError(msg)
        self.writer.writerow(row_list)
        self.file.flush()

    def close(self):
        self.file.flush()
        self.file.close()


def before(victor, loser, ballot):
    """
        Used to calculate condorcet stats. Each ballot passed through this
        function gets mapped to either
        1 (winner ranked before loser),
        0 (neither appear on ballot),
        or -1 (loser ranked before winner).
    """
    for rank in ballot:
        if rank == victor:
            return 1
        if rank == loser:
            return -1
    return 0


def remove(x, lst):
    # removes all x from list l
    return [i for i in lst if i != x]


def isInf(x):
    # checks if x is inf
    return x == float('inf')


def index_inf(lst, el):
    # return element index if in list, inf otherwise
    if el in lst:
        return lst.index(el)
    else:
        return float('inf')


def verifyDir(dir_path, make_if_missing=True, error_msg_tail='is not an existing folder'):
    """
    Check that a directory exists and if missing, either error or create it.

    :param dir_path: directory path to verify
    :param make_if_missing: if True, create directory if missing
    :param error_msg_tail: if make_if_missing is False and directory missing,
     print this error message after the dir_path.
    """
    if os.path.isdir(dir_path) is False:
        if make_if_missing:
            os.mkdir(longname(dir_path))
        else:
            print(dir_path + ' ' + error_msg_tail)
            raise RuntimeError


def flatten_list(lst):
    return [i for sublist in lst for i in sublist]


def decimal2float(stat, round_places=3):
    """Convert any decimal objects used internally into float for reporting.

    Args:
        stat (any): Any value.

    Returns:
        any type not Decimal: If the stat passed is type Decimal, it is converted to float.
    """

    if isinstance(stat, decimal.Decimal):
        return round(float(stat), round_places)
    else:
        return stat


def DL2LD(dl):
    return [dict(zip(dl, t)) for t in zip(*dl.values())]


def LD2DL(ld):
    # assumes all dicts have same keys, which they should in these use cases
    return {k: [dic[k] for dic in ld] for k in ld[0]}


def longname(path):
    return pathlib.Path('\\\\?\\' + os.fspath(path.resolve()))


def filter_bool_dict(ballots, field_name):
    val_list = ballots[field_name]
    return {split_val: [split_val == i for i in val_list] for split_val in set(val_list)}




from typing import List

import collections

import pandas as pd

# import rcv_cruncher.util as util

from rcv_cruncher.marks import BallotMarks
from rcv_cruncher.util import InactiveType


class RCV_stats:
    """
    Mixin containing all reporting stats. Can be overriden by any rcv variant.
    """

    def _exhaustion_categories(self, *, tabulation_num=1):
        """
        Returns a list with constants indicating why each ballot
        was exhausted in a single-winner rcv contest.

        Possible list values are:

        - UNDERVOTE : if the ballot was undervote, and therefore neither active nor exhaustable.

        - NOT_EXHAUSTED: if finalist was present on the ballot and was ranked higher than an exhaust condition
        (overvote or repeated_skipvotes)

        - PRETALLY_EXHAUST: if the ballot was exahusted by overvote or skipped rankings prior to being counted in the first round.

        - POSTTALLY_EXHAUSTED_BY_OVERVOTE: overvote rules apply to contest, and an overvote is encountered prior to a
        final round candidate or another exhaust condition.

        - POSTTALLY_EXHAUSTED_BY_REPEATED_SKIPVOTE: skipped rankings rules apply to contest, and two or more repeated skipped
         rankings are encountered prior to a final round candidate or another exhaust condition. The skipped rankings
          must be followed by a non-skipped ranking for this condition to apply.

        - POSTTALLY_EXHAUSTED_BY_DUPLICATE_RANKING: repeated rankings rules apply to contest, and a candidate that had already appeared
         on the ballot is encountered again prior to a final round candidate or another exhaust condition.

        - POSTTALLY_EXHAUSTED_BY_ABSTENTION: if the ballot is rank restricted, then a ballot receives this label if the final
        rank was skipped. If the ballot is not rank restricted, then all ballots that do not reach another exhaust
        condition and do not rank a final round candidate receive this label.

        - POSTTALLY_EXHAUSTED_BY_RANK_LIMIT: if the ballot is rank restricted, then a ballot recieves this label if the final
        rank was marked. If the ballot is not rank restricted, then no ballots recieve this label.

        rank restricted ballot: less than or equal to n-2 ranks, where n is number of candidates (not counting writeins).
        """

        restrictive_rank_limit = self._summary_cvr_stat_table['restrictive_rank_limit'].item()

        used_last_rank_list = self._cvr_stat_table['used_last_rank']
        final_ranks_list = self.get_final_ranks(tabulation_num=tabulation_num)
        ballot_marks_list = self.get_cvr_dict(self._contest_rule_set_name)['ballot_marks']

        why_exhaust = []
        # loop through each ballot
        for used_last_rank, final_ranks, ballot_marks in zip(used_last_rank_list, final_ranks_list, ballot_marks_list):

            # if the exhaust status is already known
            if ballot_marks.inactive_type == BallotMarks.UNDERVOTE:
                why_exhaust.append(InactiveType.UNDERVOTE)

            elif ballot_marks.inactive_type == BallotMarks.PRETALLY_EXHAUST:
                why_exhaust.append(InactiveType.PRETALLY_EXHAUST)

            # if the ballot still had some ranks at the end of tabulation
            # then it wasnt exhausted
            elif final_ranks:
                why_exhaust.append(InactiveType.NOT_EXHAUSTED)

            elif ballot_marks.inactive_type == BallotMarks.MAYBE_EXHAUSTED_BY_DUPLICATE_RANKING:
                why_exhaust.append(InactiveType.POSTTALLY_EXHAUSTED_BY_DUPLICATE_RANKING)

            elif ballot_marks.inactive_type == BallotMarks.MAYBE_EXHAUSTED_BY_REPEATED_SKIPPED_RANKING:
                why_exhaust.append(InactiveType.POSTTALLY_EXHAUSTED_BY_REPEATED_SKIPPED_RANKING)

            elif ballot_marks.inactive_type == BallotMarks.MAYBE_EXHAUSTED_BY_OVERVOTE:
                why_exhaust.append(InactiveType.POSTTALLY_EXHAUSTED_BY_OVERVOTE)

            elif restrictive_rank_limit and used_last_rank:
                why_exhaust.append(InactiveType.POSTTALLY_EXHAUSTED_BY_RANK_LIMIT)
            else:
                why_exhaust.append(InactiveType.POSTTALLY_EXHAUSTED_BY_ABSTENTION)

        return why_exhaust

    ####################
    # CONTEST INFO

    # def split_id(self):
    #     """
    #     String describing which field and value the cvr ballots were filtered on. If no filtering done, it is empty.
    #     """
    #     return self._split_id

    # def split_field(self):
    #     return self._split_field

    # def split_value(self):
    #     return self._split_value.replace(":", "_").replace("/", "_").replace("\\", "_").replace(" ", "_").replace("-", "_")

    # def file_stub(self, *, tabulation_num=None):

    #     stub = ""
    #     if not self.split_id():
    #         stub += self.unique_id()
    #     else:
    #         stub += self.split_id()

    #     if tabulation_num:
    #         stub += '_tab-' + str(tabulation_num)

    #     return stub

    ####################
    # OUTCOME STATS

    def _winner(self, tabulation_num=1):
        '''
        The winner(s) of the election.
        '''
        return ", ".join([str(w).title() for w in self._tabulation_winner(tabulation_num=tabulation_num)])

    def _all_winners(self):
        """
        Return contest winner names in order of election.
        """
        # accumulate winners across tabulations
        winners = []
        for i in range(1, self.n_tabulations() + 1):
            winners += self._tabulation_winner(tabulation_num=i)
        return winners

    def _tabulation_winner(self, tabulation_num=1):
        """
        Return winners from tabulation.
        """
        elected_candidates = [d for d in self.get_candidate_outcomes(tabulation_num=tabulation_num)
                              if d['round_elected'] is not None]
        return [d['name'] for d in sorted(elected_candidates, key=lambda x: x['round_elected'])]

    def _condorcet(self, tabulation_num=1):
        '''
        Is the winner the condorcet winner?
        The condorcet winner is the candidate that would win a 1-on-1 election versus
        any other candidate in the election. Note that this calculation depends on
        jurisdiction dependant rule variations.

        In the case of multi-winner elections, this result will only pertain to the first candidate elected.
        '''

        cands = self._contest_candidates
        if len(cands.unique_candidates) == 1:
            return True

        winner = self._tabulation_winner(tabulation_num=tabulation_num)[0]
        losers = [cand for cand in cands.unique_candidates if cand != winner]

        net = collections.Counter()
        for b in self._contest_cvr_ld:
            for loser in losers:

                # does winner or loser appear first on this ballot?
                ballot_contrib = 0
                for mark in b['ballot_marks'].marks:
                    if mark == winner:
                        ballot_contrib = 1
                        break
                    if mark == loser:
                        ballot_contrib = -1
                        break

                # accumulate
                net.update({loser: ballot_contrib})

        # any negative net values indicate a head-to-head where contest winner loses
        if min(net.values()) > 0:
            return True
        else:
            return False

    def _come_from_behind(self, tabulation_num=1):
        """
        "yes" if rcv winner is not first round leader, else "no".

        In the case of multi-winner elections, this result will only pertain to the first candidate elected.
        """
        if self._first_round_winner_place(tabulation_num=tabulation_num) != 1:
            return True
        else:
            return False

    def _final_round_active_votes(self, tabulation_num=1):
        '''
        The number of votes that were awarded to any candidate in the final round. (weighted)
        '''
        n_rounds = self.n_rounds(tabulation_num=tabulation_num)
        tally_dict = self.get_round_tally_dict(n_rounds, tabulation_num=tabulation_num)
        return sum(tally_dict.values())

    def _final_round_winner_percent(self, tabulation_num=1):
        '''
        The percent of votes for the winner in the final round.
        If more than one winner, return the final round count for the first winner elected. (weighted)
        '''
        n_rounds = self.n_rounds(tabulation_num=tabulation_num)
        tally_dict = self.get_round_tally_dict(n_rounds, tabulation_num=tabulation_num)
        winner = self._tabulation_winner(tabulation_num=tabulation_num)
        return (tally_dict[winner[0]] / sum(tally_dict.values())) * 100

    def _final_round_winner_vote(self, tabulation_num=1):
        '''
        The percent of votes for the winner in the final round.
        If more than one winner, return the final round count for the first winner elected. (weighted)
        '''
        n_rounds = self.n_rounds(tabulation_num=tabulation_num)
        tally_dict = self.get_round_tally_dict(n_rounds, tabulation_num=tabulation_num)
        winner = self._tabulation_winner(tabulation_num=tabulation_num)
        return tally_dict[winner[0]]

    def _final_round_winner_votes_over_first_round_active(self, tabulation_num=1):
        '''
        The number of votes the winner receives in the final round divided by the
        number of valid votes in the first round. Reported as percentage.

        If more than one winner, return the final round count for the first winner elected. (weighted)
        '''
        first_round_active_votes = sum(self.get_round_tally_dict(1, tabulation_num=tabulation_num).values())
        return (self._final_round_winner_vote(tabulation_num=tabulation_num) / first_round_active_votes) * 100

    def _first_round_winner_place(self, tabulation_num=1):
        '''
        In terms of first round votes, what place the eventual winner came in.
        In the case of multi-winner elections, this result will only pertain to the first candidate elected.
        '''
        winner = self._tabulation_winner(tabulation_num=tabulation_num)[0]
        tally_tuple = self.get_round_tally_tuple(1, tabulation_num=tabulation_num,
                                                 only_round_active_candidates=True, desc_sort=True)

        # account for ties
        winner_place = None
        for order_rank, unique_tally_val in enumerate(sorted(set(tally_tuple[1]), reverse=True), start=1):
            for cand, tally in zip(*tally_tuple):
                if winner == cand and tally == unique_tally_val:
                    winner_place = order_rank

        return winner_place

    def _first_round_winner_percent(self, tabulation_num=1):
        '''
        The percent of votes for the winner in the first round.
        In the case of multi-winner elections, this result will only pertain to the first candidate elected. (weighted)
        '''
        winner = self._tabulation_winner(tabulation_num=tabulation_num)
        tally_dict = self.get_round_tally_dict(1, tabulation_num=tabulation_num, only_round_active_candidates=True)
        return tally_dict[winner[0]] / sum(tally_dict.values()) * 100

    def _first_round_winner_vote(self, tabulation_num=1):
        '''
        The number of votes for the winner in the first round.
        In the case of multi-winner elections, this result will only pertain to the first candidate elected. (weighted)
        '''
        winner = self._tabulation_winner(tabulation_num=tabulation_num)
        tally_dict = self.get_round_tally_dict(1, tabulation_num=tabulation_num, only_round_active_candidates=True)
        return tally_dict[winner[0]]

    def _number_of_winners(self):
        """
        Number of winners a contest had.
        """
        return len(self._all_winners())

    def _ranked_winner(self, tabulation_num=1):
        """
        Number of ballots with a non-overvote mark for the winner. (weighted) (filtered)
        """
        winners = self._tabulation_winner(tabulation_num=tabulation_num)
        winner_marked = [bool(set(winners).intersection(b['ballot_marks'].unique_marks)) for b in self._contest_cvr_ld]
        return sum(b['weight'] for flag, b in zip(winner_marked, self._contest_cvr_ld) if flag)

    def _win_threshold(self, tabulation_num=1):
        """
        Election threshold, if static, otherwise NA
        """
        return self.get_win_threshold(tabulation_num=tabulation_num)

    def _winners_consensus_value(self, tabulation_num=1):
        '''
        The percentage of valid first round votes that rank any winner in the top 3.
        '''
        first_round_active_votes = sum(self.get_round_tally_dict(1, tabulation_num=tabulation_num).values())
        return (self._winner_in_top_3(tabulation_num=tabulation_num) / first_round_active_votes) * 100

    def _winner_in_top_3(self, tabulation_num=1):
        """
        Number of ballots that ranked any winner in the top 3 ranks. (weighted)
        """
        winner = self._tabulation_winner(tabulation_num=tabulation_num)
        top3 = [b['ballot_marks'].marks[:min(3, len(b['ballot_marks'].marks))] for b in self._contest_cvr_ld]
        top3_check = [bool(set(winner).intersection(b)) for b in top3]
        return sum(b['weight'] for flag, b in zip(top3_check, self._contest_cvr_ld) if flag)

    def _compute_contest_stat_table(self):

        cvr = self.get_cvr_dict(self._contest_rule_set_name)

        df = pd.DataFrame()

        # ADD WEIGHTS
        df['weight'] = cvr['weight']
        for iTab in range(1, self._tab_num+1):
            df[f'final_weight{iTab}'] = self.get_final_weights(tabulation_num=iTab)

        # EXHAUSTION STATS
        for iTab in range(1, self._tab_num+1):
            exhaust_type = self._exhaustion_categories(tabulation_num=iTab)
            df[f'exhaust_type{iTab}'] = pd.Series(exhaust_type, dtype='category')

        for iTab in range(1, self._tab_num+1):
            exhaust_type_str = f'exhaust_type{iTab}'
            df[f'pretally_exhausted{iTab}'] = df[exhaust_type_str].eq(
                InactiveType.PRETALLY_EXHAUST)
            df[f'posttally_exhausted_by_overvote{iTab}'] = df[exhaust_type_str].eq(
                InactiveType.POSTTALLY_EXHAUSTED_BY_OVERVOTE)
            df[f'posttally_exhausted_by_repeated_skipped_rankings{iTab}'] = df[exhaust_type_str].eq(
                InactiveType.POSTTALLY_EXHAUSTED_BY_REPEATED_SKIPPED_RANKING)
            df[f'posttally_exhausted_by_abstention{iTab}'] = df[exhaust_type_str].eq(
                InactiveType.POSTTALLY_EXHAUSTED_BY_ABSTENTION)
            df[f'posttally_exhausted_by_rank_limit{iTab}'] = df[exhaust_type_str].eq(
                InactiveType.POSTTALLY_EXHAUSTED_BY_RANK_LIMIT)
            df[f'posttally_exhausted_by_duplicate_rankings{iTab}'] = df[exhaust_type_str].eq(
                InactiveType.POSTTALLY_EXHAUSTED_BY_DUPLICATE_RANKING)

            all_posttally_conditions = [
                f'posttally_exhausted_by_overvote{iTab}',
                f'posttally_exhausted_by_repeated_skipped_rankings{iTab}',
                f'posttally_exhausted_by_abstention{iTab}',
                f'posttally_exhausted_by_rank_limit{iTab}',
                f'posttally_exhausted_by_duplicate_rankings{iTab}'
            ]
            df['posttally_exhausted'+str(iTab)] = df[all_posttally_conditions].any(axis='columns')

        self._contest_stat_table = df

    def _compute_summary_contest_stat_tables(self) -> None:

        tabulation_stats = []

        for iTab in range(1, self._tab_num+1):

            s = pd.Series()

            s['rcv_type'] = self.__class__.__name__

            exhaust_on_overvote = self._rule_sets[self._contest_rule_set_name]['exhaust_on_overvote_marks']
            exhaust_on_repeated_skipped = self._rule_sets[self._contest_rule_set_name]['exhaust_on_repeated_skipped_marks']
            exhaust_on_duplicate = self._rule_sets[self._contest_rule_set_name]['exhaust_on_duplicate_candidate_marks']
            combine_writeins = self._rule_sets[self._contest_rule_set_name]['combine_writein_marks']
            exclude_writeins = self._rule_sets[self._contest_rule_set_name]['exclude_writein_marks']
            treat_writeins = self._rule_sets[self._contest_rule_set_name]['treat_combined_writeins_as_exhaustable_duplicates']

            s['exhaust_on_overvote_marks'] = exhaust_on_overvote
            s['exhaust_on_repeated_skipped_marks'] = exhaust_on_repeated_skipped
            s['exhaust_on_duplicate_candidate_marks'] = exhaust_on_duplicate
            s['combine_writein_marks'] = combine_writeins,
            s['exclude_writein_marks'] = exclude_writeins,
            s['treat_combined_writeins_as_exhaustable_duplicates'] = treat_writeins

            s['number_of_winners'] = self._number_of_winners()

            s['tabulation_num'] = iTab
            s['winner'] = self._winner(tabulation_num=iTab)
            s['number_of_rounds'] = self.n_rounds(tabulation_num=iTab)
            s['winners_consensus_value'] = self._winners_consensus_value(tabulation_num=iTab)

            first_round_active_votes = sum(self.get_round_tally_dict(1, tabulation_num=iTab).values())
            s['first_round_active_votes'] = first_round_active_votes

            final_round_active_votes = sum(self.get_round_tally_dict(s['number_of_rounds'], tabulation_num=iTab).values())
            s['final_round_active_votes'] = final_round_active_votes

            weight = self._contest_stat_table[f'final_weight{iTab}']

            pretally = sum(self._contest_stat_table[f'pretally_exhausted{iTab}'] * weight)
            s['total_pretally_exhausted'] = pretally

            posttally = sum(self._contest_stat_table[f'posttally_exhausted{iTab}'] * weight)
            s['total_posttally_exhausted'] = posttally

            posttally_overvote = sum(self._contest_stat_table[f'posttally_exhausted_by_overvote{iTab}'] * weight)
            s['total_posttally_exhausted_by_overvote'] = posttally_overvote

            posttally_skipped = sum(self._contest_stat_table[f'posttally_exhausted_by_repeated_skipped_rankings{iTab}'] * weight)
            s['total_posttally_exhausted_by_skipped_rankings'] = posttally_skipped

            posttally_abstention = sum(self._contest_stat_table[f'posttally_exhausted_by_abstention{iTab}'] * weight)
            s['total_posttally_exhausted_by_abstention'] = posttally_abstention

            posttally_rank_limit = sum(self._contest_stat_table[f'posttally_exhausted_by_rank_limit{iTab}'] * weight)
            s['total_posttally_exhausted_by_rank_limit'] = posttally_rank_limit

            posttally_duplicate = sum(self._contest_stat_table[f'posttally_exhausted_by_duplicate_rankings{iTab}'] * weight)
            s['total_posttally_exhausted_by_duplicate_rankings'] = posttally_duplicate

            if len(self._tabulation_winner(tabulation_num=iTab)) == 1:

                s['first_round_winner_vote'] = self._first_round_winner_vote(tabulation_num=iTab)
                s['final_round_winner_vote'] = self._final_round_winner_vote(tabulation_num=iTab)
                s['first_round_winner_percent'] = self._first_round_winner_percent(tabulation_num=iTab)
                s['final_round_winner_percent'] = self._final_round_winner_percent(tabulation_num=iTab)
                s['first_round_winner_place'] = self._first_round_winner_place(tabulation_num=iTab)
                s['condorcet'] = self._condorcet(tabulation_num=iTab)
                s['come_from_behind'] = self._come_from_behind(tabulation_num=iTab)
                s['ranked_winner'] = self._ranked_winner(tabulation_num=iTab)

                final_over_first = self._final_round_winner_votes_over_first_round_active(tabulation_num=iTab)
                s['final_round_winner_votes_over_first_round_active'] = final_over_first

                s['win_threshold'] = None

            else:

                s['first_round_winner_vote'] = None
                s['final_round_winner_vote'] = None
                s['first_round_winner_percent'] = None
                s['final_round_winner_percent'] = None
                s['first_round_winner_place'] = None
                s['final_round_winner_votes_over_first_round_active'] = None
                s['condorcet'] = None
                s['come_from_behind'] = None
                s['ranked_winner'] = None

                s['win_threshold'] = self.get_win_threshold(tabulation_num=iTab)

            tabulation_stats.append(s.to_frame().transpose())

        self._summary_contest_stat_tables = tabulation_stats

    def _compute_contest_split_stats(self, split_filter: List[bool]) -> pd.DataFrame:

        tabulation_split_stats = []

        filtered_stat_table = self._contest_stat_table.loc[split_filter, :]

        for iTab in range(1, self._tab_num+1):

            s = pd.Series()

            weight = filtered_stat_table[f'final_weight{iTab}']

            pretally = sum(filtered_stat_table[f'pretally_exhausted{iTab}'] * weight)
            s['split_total_pretally_exhausted'] = pretally

            posttally = sum(filtered_stat_table[f'posttally_exhausted{iTab}'] * weight)
            s['split_total_posttally_exhausted'] = posttally

            posttally_overvote = sum(filtered_stat_table[f'posttally_exhausted_by_overvote{iTab}'] * weight)
            s['split_total_posttally_exhausted_by_overvote'] = posttally_overvote

            posttally_skipped = sum(filtered_stat_table[f'posttally_exhausted_by_repeated_skipped_rankings{iTab}'] * weight)
            s['split_total_posttally_exhausted_by_skipped_rankings'] = posttally_skipped

            posttally_abstention = sum(filtered_stat_table[f'posttally_exhausted_by_abstention{iTab}'] * weight)
            s['split_total_posttally_exhausted_by_abstention'] = posttally_abstention

            posttally_rank_limit = sum(filtered_stat_table[f'posttally_exhausted_by_rank_limit{iTab}'] * weight)
            s['split_total_posttally_exhausted_by_rank_limit'] = posttally_rank_limit

            posttally_duplicate = sum(filtered_stat_table[f'posttally_exhausted_by_duplicate_rankings{iTab}'] * weight)
            s['split_total_posttally_exhausted_by_duplicate_rankings'] = posttally_duplicate

            tabulation_split_stats.append(s.to_frame().transpose())

        return tabulation_split_stats

    def _compute_summary_contest_split_stat_tables(self) -> None:

        if not self._split_filter_dict:
            return

        split_tabulation_stat_df_list = [[]] * self._tab_num

        for field in self._split_filter_dict:
            field_clean = self._clean_string(field)

            for unique_val in self._split_filter_dict[field]:
                val_clean = self._clean_string(unique_val)
                split_id = field_clean + "-" + val_clean

                split_id_df = pd.DataFrame({
                    'split_field': [field],
                    'split_value': [unique_val],
                    'split_id': [split_id]
                    })
                split_stat_df_list = self._compute_contest_split_stats(self._split_filter_dict[field][unique_val])

                for split_stat_df_idx, split_stat_df in enumerate(split_stat_df_list):
                    split_tabulation_stat_df_list[split_stat_df_idx].append(
                        pd.concat([split_id_df, split_stat_df], axis='columns'))

        summary_contest_split_stat_tables = [pd.concat(split_stat_list, axis=0, ignore_index=True, sort=False)
                                             for split_stat_list in split_tabulation_stat_df_list]
        self._summary_contest_split_stat_tables = summary_contest_split_stat_tables




# import json
# import os
# import pathlib
# import re
# import shutil
# import pkg_resources
# import collections
# import datetime
# import abc

# import pandas as pd
# import tqdm

# import rcv_cruncher.parsers as parsers
# import rcv_cruncher.rcv_base as rcv_base
# import rcv_cruncher.rcv_variants as rcv_variants
# import rcv_cruncher.util as util
# import rcv_cruncher.write_out as write_out


# # read functions in parsers and rcv_variants
# rcv_dict = rcv_variants.get_rcv_dict()
# parser_dict = parsers.get_parser_dict()


# @staticmethod
# def new_rcv_contest(contest_dict: Dict) -> Type[RCV]:
#     """
#     Pass in a dictionary and run the constructor function stored within it
#     """
#     param_names = inspect.signature(contest_dict['rcv_type'].__init__)
#     param_dict = {k: v for k, v in contest_dict.items() if k in param_names}
#     return contest_dict['rcv_type'](**param_dict)


# def dummy(*args):
#     pass


# # typecast functions
# def cast_str(s):
#     """
#     If string-in-string '"0006"', evaluate to '0006'
#     If 'None', return None (since this string cannot be evaluated)
#     else, return str() result
#     """
#     if (s[0] == '"' and s[-1] == '"') or (s[0] == "'" and s[-1] == "'"):
#         return eval(s)
#     elif s == 'None':
#         return None
#     else:
#         return str(s)


# def cast_int(s):
#     if isinstance(s, int):
#         return s
#     return int(s)


# def cast_bool(s):
#     if isinstance(s, bool):
#         return s
#     return eval(s.title())


# def cast_list(lst):
#     if lst == "":
#         return []
#     return lst.strip('\n').split(",")


# def cast_dict(dct):
#     dct_return = {}
#     if dct == "":
#         return dct_return
#     comma_split = dct.strip("\n").split(",")
#     for i in comma_split:
#         equal_split = i.split("=")
#         dct_return.update({equal_split[0]: "=".join(equal_split[1:])})
#     return dct_return


# def cast_func(s):

#     if s in rcv_dict and s in parser_dict:
#         raise RuntimeError('(developer error) An rcv variant class and a parser function share the same name. Make them unique.')

#     if s in rcv_dict:
#         return rcv_dict[s]

#     if s in parser_dict:
#         return parser_dict[s]


# # other helpers
# # def dop(ctx):
# #     return '_'.join([ctx['year'], ctx['place'], ctx['office']])

# def unique_id(ctx):
#     filled_date = "/".join([piece if len(piece) > 1 else "0" + piece for piece in ctx['date'].split("/")])
#     pieces = [ctx['jurisdiction'], filled_date, ctx['office']]
#     cleaned_pieces = [re.sub('[^0-9a-zA-Z_]+', '', x) for x in pieces]
#     return "_".join(cleaned_pieces)

# def read_contest_set(contest_set_path, override_cvr_root_dir=None):

#     contest_set_path = pathlib.Path(contest_set_path)

#     # assemble typecast funcs
#     cast_dict = {'str': cast_str, 'int': cast_int, 'dict': cast_dict,
#                  'bool': cast_bool, 'func': cast_func, 'list': cast_list}

#     # settings/defaults
#     contest_set_settings_fpath = f'{os.path.dirname(__file__)}/contest_set_settings.json'
#     if os.path.isfile(contest_set_settings_fpath) is False:
#         raise RuntimeError(f'(developer error) Looking for contest_set_settings.json. Not a valid file path: {contest_set_settings_fpath}')

#     with open(contest_set_settings_fpath) as contest_set_settings_file:
#         contest_set_settings = json.load(contest_set_settings_file)

#     run_config_settings_fpath = f'{os.path.dirname(__file__)}/run_config_settings.json'
#     if os.path.isfile(run_config_settings_fpath) is False:
#         raise RuntimeError(f'(developer error) Looking for run_config_settings.json. Not a valid file path: {run_config_settings_fpath}')

#     with open(run_config_settings_fpath) as run_config_settings_file:
#         run_config_settings = json.load(run_config_settings_file)

#     # read run_config.txt
#     run_config_fpath = contest_set_path / 'run_config.txt'
#     if os.path.isfile(run_config_fpath) is False:
#         raise RuntimeError(f'not a valid file path: {run_config_fpath}')

#     run_config = {}
#     with open(run_config_fpath) as run_config_file:
#         for line_num, l in enumerate(run_config_file, start=1):

#             l_splits = l.strip('\n').split("=")
#             l_splits = [s.strip() for s in l_splits]

#             if len(l_splits) < 2:
#                 continue

#             input_option = l_splits[0]
#             input_value = l_splits[1]

#             if input_option not in run_config_settings:
#                 continue

#             if run_config_settings[input_option]['type'] == "bool":
#                 if input_value.title() != "True" and input_value.title() != "False":
#                     raise RuntimeError(f'invalid value ({input_value}) provided in run_config.txt'
#                                        f'on line {line_num} for option "{l_splits[0]}". Must be "true" or "false".')
#                 input_value = eval(input_value.title())

#             run_config.update({input_option: input_value})

#     # add in defaults for missing options
#     for field in run_config_settings:
#         if field not in run_config:
#             run_config.update({field: run_config_settings[field]['default']})

#     run_config['cvr_path_root'] = pathlib.Path(run_config['cvr_path_root'])

#     # read contest_set.csv
#     contest_set_fpath = contest_set_path / 'contest_set.csv'
#     if os.path.isfile(contest_set_fpath) is False:
#         raise RuntimeError(f'not a valid file path: {contest_set_fpath}')

#     contest_set_df = pd.read_csv(contest_set_fpath, dtype=object)

#     # add in default values for missing columns
#     for setting in contest_set_settings:
#         if setting not in contest_set_df.columns:
#             contest_set_df[setting] = contest_set_settings[setting]['default']

#     # fill in na values with defaults and evaluate column, if indicated
#     for col in contest_set_df:

#         if col not in contest_set_settings:
#             print(f'info -- "{col}" is an unrecognized column in contest_set.csv, it will be ignored.')
#         else:
#             contest_set_df[col] = contest_set_df[col].fillna(contest_set_settings[col]['default'])
#             contest_set_df[col] = [cast_dict[contest_set_settings[col]['type']](i) for i in contest_set_df[col].tolist()]

#     contest_set_df['contest_set_path'] = contest_set_path

#     # convert df to listOdicts, one dict per row
#     competitions = contest_set_df.to_dict('records')

#     # add dop, uid, cvr_path_root
#     for d in competitions:

#         if override_cvr_root_dir:
#             d['cvr_path'] = pathlib.Path(override_cvr_root_dir) / d['cvr_path']
#         else:
#             d['cvr_path'] = run_config['cvr_path_root'] / d['cvr_path']

#         d['parser_args'] = {'cvr_path': d['cvr_path']}
#         d['parser_args'].update(d['extra_parser_args'])

#         if d['candidate_map']:
#             d['candidate_map'] = run_config['cvr_path_root'] / d['candidate_map']

#         d['uid'] = util.unique_id(d)

#     # remove contest that should be ignored
#     competitions = [comp for comp in competitions if not comp.get('ignore_contest', True)]

#     # remove contests with invalid parser, print warning.
#     no_parser = []
#     for comp in competitions:
#         if comp['parser'] == dummy:
#             no_parser.append(comp)

#     no_parser_uniqueID = [comp['uid'] for comp in no_parser]
#     valid_competitions = [comp for comp in competitions if comp['uid'] not in no_parser_uniqueID]

#     if no_parser:
#         print(f'info -- {len(no_parser)} contests in contest_set.csv did not include a valid cvr parser field. They will be ignored.')

#     # store file locations
#     run_config['contest_set_file_path'] = contest_set_fpath
#     run_config['run_config_file_path'] = run_config_fpath

#     return valid_competitions, run_config


# def write_aggregated_stats(results_dir,
#                            output_config,
#                            rcv_group_stats_df_dict,
#                            rcv_variant_stats_df_dict,
#                            candidate_details_dfs,
#                            quiet=False):

#     if not quiet:
#         print("####################")
#         print("Write stored results")

#     if output_config.get('per_rcv_group_stats', False):

#         if not quiet:
#             print("write group stats ...")

#         for group in rcv_group_stats_df_dict:
#             if rcv_group_stats_df_dict[group]:
#                 df = pd.concat(rcv_group_stats_df_dict[group], axis=0, ignore_index=True, sort=False)
#                 df.to_csv(util.longname(results_dir / f'group_{group}.csv'), index=False)

#     if output_config.get('per_rcv_group_stats_fvDBfmt', False):

#         if not quiet:
#             print("write group stats in fvDB order ...")

#         for group in rcv_group_stats_df_dict:
#             format_fpath = f"extra/fv_db_format/{group}_columns.csv"
#             if rcv_group_stats_df_dict[group] and os.path.isfile(format_fpath):

#                 # read in column order
#                 fmt_df = pd.read_csv(format_fpath)
#                 fmt_order = fmt_df['cruncher_col'].tolist()

#                 df = pd.concat(rcv_group_stats_df_dict[group], axis=0, ignore_index=True, sort=False)
#                 df = df.reindex(fmt_order, axis=1)
#                 df.to_csv(util.longname(results_dir / f'group_{group}_fvDBfmt.csv'), index=False)

#     if output_config.get('per_rcv_type_stats', False):

#         if not quiet:
#             print("Write tabulation stats ...")

#         for variant in rcv_variant_stats_df_dict:
#             if rcv_variant_stats_df_dict[variant]:
#                 df = pd.concat(rcv_variant_stats_df_dict[variant], axis=0, ignore_index=True, sort=False)
#                 df.to_csv(util.longname(results_dir / f'{variant}.csv'), index=False)

#     if output_config.get('candidate_details', False) and candidate_details_dfs:

#         if not quiet:
#             print("Write candidate details ...")

#         df = pd.concat(util.flatten_list(candidate_details_dfs), axis=0, ignore_index=True, sort=False)
#         df.to_csv(util.longname(results_dir / 'candidate_details.csv'), index=False)


# class Steps(abc.ABC):

#     def __init__(self, contest, output_config, converted_cvr_dir, results_dir, pbar_desc):

#         self.contest = contest
#         self.output_config = output_config
#         self.converted_cvr_dir = converted_cvr_dir
#         self.results_dir = results_dir
#         self.pbar_desc = pbar_desc
#         self.error_log_writers = []

#         self.state_data = {
#             'n_errors': 0,
#             'rcv_obj': None
#         }
#         self.steps = {}

#     def update_state(self, dct):
#         self.state_data.update(dct)

#     def update_pbar_desc(self, desc):
#         self.pbar_desc = desc

#     def update_error_log_writers(self, writers_list):
#         self.error_log_writers = writers_list if isinstance(writers_list, list) else [writers_list]

#     def refresh_steps(self):

#         cache_keys = ['success', 'order']
#         cache = collections.defaultdict(dict)
#         for k1 in self.steps:
#             for k2 in cache_keys:
#                 if k2 in self.steps[k1]:
#                     cache[k1].update({k2: self.steps[k1][k2]})

#         self.steps = self.generate_steps()

#         for k in self.steps:
#             for cache_key in cache_keys:
#                 if cache_key in cache[k]:
#                     self.steps[k][cache_key] = cache[k][cache_key]

#     @abc.abstractmethod
#     def generate_steps(self):
#         pass

#     def n_steps(self):
#         return len(self.steps.keys())

#     def next_step(self):

#         remaining_steps = [
#             (k, step) for k, step in self.steps.items()
#             if step['success'] is None  # step not attempted yet
#             and step['condition']  # step conditions are met
#             and not any(self.steps[dep_k]['success'] is False for dep_k in step['fail_with'])  # all step dependencies are met
#             and all(self.steps[dep_k]['success'] for dep_k in step['depends_on'])  # all step dependencies are met
#         ]

#         if not remaining_steps:
#             return False
#         else:
#             return remaining_steps[0]

#     def run_steps(self):

#         self.state_data['n_errors'] = 0

#         # init
#         self.refresh_steps()
#         for step_num, k in enumerate(self.steps, start=1):
#             self.steps[k]['success'] = None
#             self.steps[k]['order'] = step_num

#         step_reached = 0

#         pbar = tqdm.tqdm(total=self.n_steps(), bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt}{postfix}', colour='GREEN')
#         pbar.set_description(self.pbar_desc)

#         next_step = self.next_step()
#         while next_step:

#             step_name, step_details = next_step

#             try:

#                 pbar.set_postfix_str(step_name)

#                 if step_details['return_key']:
#                     self.state_data.update({
#                         step_details['return_key']:
#                         step_details['f'](*step_details['args'])
#                         })
#                 else:
#                     step_details['f'](*step_details['args'])

#             except Exception as e:

#                 self.steps[step_name]['success'] = False

#                 for writer in self.error_log_writers:
#                     writer.write([self.contest['uid'], self.contest['split_id'], step_name, repr(e)])
#                 self.state_data['n_errors'] += 1

#             else:
#                 self.steps[step_name]['success'] = True

#             finally:
#                 pbar.update(step_details['order']-step_reached)
#                 step_reached = step_details['order']

#             self.refresh_steps()
#             next_step = self.next_step()

#         pbar.set_postfix_str('complete')
#         pbar.close()

#     def return_results(self):
#         return self.state_data


# class CrunchSteps(Steps):

#     def generate_steps(self):

#         return collections.OrderedDict([
#             ('convert_cvr_rank', {
#                 'f': write_out.write_converted_cvr,
#                 'args': [self.contest, self.converted_cvr_dir, 'rank'],
#                 'condition': self.output_config.get('convert_cvr_rank_format'),
#                 'fail_with': [],
#                 'depends_on': [],
#                 'return_key': None
#             }),
#             ('convert_cvr_candidate', {
#                 'f': write_out.write_converted_cvr,
#                 'args': [self.contest, self.converted_cvr_dir, 'candidate'],
#                 'condition': self.output_config.get('convert_cvr_candidate_format'),
#                 'fail_with': [],
#                 'depends_on': [],
#                 'return_key': None
#             }),
#             ('tabulate', {
#                 'f': rcv_base.RCV.run_rcv,
#                 'args': [self.contest],
#                 'condition': True,
#                 'depends_on': [],
#                 'fail_with': [],
#                 'return_key': 'rcv_obj'
#             }),
#             ('tabulation_stats', {
#                 'f': rcv_base.RCV.get_tabulation_stats_df,
#                 'args': [self.state_data['rcv_obj']],
#                 'condition': self.output_config.get('per_rcv_type_stats'),
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': 'tabulation_stats_df'
#             }),
#             ('rcv_variant', {
#                 'f': rcv_base.RCV.get_variant_name,
#                 'args': [self.state_data['rcv_obj']],
#                 'condition': True,
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': 'variant'
#             }),
#             ('contest_stats', {
#                 'f': rcv_base.RCV.get_contest_stats_df,
#                 'args': [self.state_data['rcv_obj']],
#                 'condition': self.output_config.get('per_rcv_group_stats'),
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': 'contest_stats_df'
#             }),
#             ('rcv_group', {
#                 'f': rcv_base.RCV.get_variant_group,
#                 'args': [self.state_data['rcv_obj']],
#                 'condition': True,
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': 'variant_group'
#             }),
#             ('candidate_details', {
#                 'f': write_out.prepare_candidate_details,
#                 'args': [self.state_data['rcv_obj']],
#                 'condition': self.output_config.get('candidate_details'),
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': 'candidate_details'
#             }),
#             ('round_by_round', {
#                 'f': write_out.write_rcv_rounds,
#                 'args': [self.state_data['rcv_obj'], self.results_dir],
#                 'condition': self.output_config.get('round_by_round'),
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': None
#             }),
#             ('ballot_stats_debug', {
#                 'f': write_out.write_ballot_debug_info,
#                 'args': [self.state_data['rcv_obj'], self.results_dir],
#                 'condition': self.output_config.get('ballot_stats_debug'),
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': None
#             }),
#             ('cvr_ballot_allocation_rank', {
#                 'f': write_out.write_converted_cvr_annotated,
#                 'args': [self.state_data['rcv_obj'], self.results_dir, 'rank'],
#                 'condition': self.output_config.get('cvr_ballot_allocation_rank_format'),
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': None
#             }),
#             ('cvr_ballot_allocation_candidate', {
#                 'f': write_out.write_converted_cvr_annotated,
#                 'args': [self.state_data['rcv_obj'], self.results_dir, 'candidate'],
#                 'condition': self.output_config.get('cvr_ballot_allocation_candidate_format'),
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': None
#             }),
#             ('first_choice_to_finalist', {
#                 'f': write_out.write_first_to_finalist_tables,
#                 'args': [self.state_data['rcv_obj'], self.results_dir],
#                 'condition': self.output_config.get('first_choice_to_finalist'),
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': None
#             }),
#             ('condorcet', {
#                 'f': write_out.write_condorcet_tables,
#                 'args': [self.contest, self.results_dir],
#                 'condition': self.output_config.get('condorcet'),
#                 'depends_on': [],
#                 'fail_with': [],
#                 'return_key': None
#             }),
#             ('first_second_choices', {
#                 'f': write_out.write_first_second_tables,
#                 'args': [self.contest, self.results_dir],
#                 'condition': self.output_config.get('first_second_choices'),
#                 'depends_on': [],
#                 'fail_with': [],
#                 'return_key': None
#             }),
#             ('cumulative_rankings', {
#                 'f': write_out.write_cumulative_ranking_tables,
#                 'args': [self.contest, self.results_dir],
#                 'condition': self.output_config.get('cumulative_rankings'),
#                 'depends_on': [],
#                 'fail_with': [],
#                 'return_key': None
#             }),
#             ('rank_usage', {
#                 'f': write_out.write_rank_usage_tables,
#                 'args': [self.contest, self.results_dir],
#                 'condition': self.output_config.get('rank_usage'),
#                 'depends_on': [],
#                 'fail_with': [],
#                 'return_key': None
#             }),
#             ('crossover_support', {
#                 'f': write_out.write_opponent_crossover_tables,
#                 'args': [self.contest, self.results_dir],
#                 'condition': self.output_config.get('crossover_support'),
#                 'depends_on': [],
#                 'fail_with': [],
#                 'return_key': None
#             }),
#             ('split_stats', {
#                 'f': rcv_base.RCV.get_split_stats,
#                 'args': [self.state_data['rcv_obj']],
#                 'condition': True,
#                 'depends_on': ['tabulate'],
#                 'fail_with': [],
#                 'return_key': 'split_stats'
#             })
#         ])


# def init_error_logger(results_dir):
#     header_list = ['contest', 'split_id', 'cruncher_step', 'message']
#     error_log_path = results_dir / 'error_log.csv'
#     return util.CSVLogger(error_log_path, header_list)


# def write_input_dir(results_dir, output_config, start_time, end_time):

#     # copy input files
#     result_log_dir = results_dir / 'inputs'
#     util.verifyDir(result_log_dir)

#     pkg_version = pkg_resources.require("rcv_cruncher")[0].version
#     pkg_url = "https://github.com/fairvotereform/rcv_cruncher"
#     with open(result_log_dir / 'pkg_info.txt', 'w') as pkg_info:
#         pkg_info.write(f'version: {pkg_version}\n')
#         pkg_info.write(f'github: {pkg_url}\n')
#         pkg_info.write(f'start_time: {start_time.strftime("%Y-%m-%d %H:%M:%S")}\n')
#         pkg_info.write(f'end_time: {end_time.strftime("%Y-%m-%d %H:%M:%S")}')

#     log_output_config = result_log_dir / 'output_config.csv'
#     shutil.copy2(output_config['run_config_file_path'], log_output_config)

#     log_contest_set_fname = result_log_dir / 'contest_set.csv'
#     shutil.copy2(output_config['contest_set_file_path'], log_contest_set_fname)


# def crunch_contest_set(contest_set, output_config, path_to_output, fresh_output=False):

#     start_time = datetime.datetime.now()

#     ##################
#     # OUTPUT PATHS
#     path_to_output = pathlib.Path(path_to_output)

#     # cvrs from path_to_cvr used in tabulation will be converted and output here
#     converted_cvr_dir = path_to_output / 'converted_cvr'
#     if fresh_output and converted_cvr_dir.exists():
#         print('deleting existing converted_cvr directory...')
#         shutil.rmtree(util.longname(converted_cvr_dir))
#     util.verifyDir(util.longname(converted_cvr_dir))

#     # various tabulation stats will be output here
#     results_dir = path_to_output / 'results'
#     if fresh_output and results_dir.exists():
#         print('deleting existing results directory...')
#         shutil.rmtree(util.longname(results_dir))
#     util.verifyDir(util.longname(results_dir))

#     #########################
#     # SOME RESULTS CONTAINERS

#     candidate_details_dfs = []
#     rcv_variant_stats_df_dict = {variant_name: [] for variant_name in rcv_variants.get_rcv_dict().keys()}
#     rcv_group_stats_df_dict = {variant_group: [] for variant_group in
#                                set(g.variant_group() for g in rcv_variants.get_rcv_dict().values())}

#     allsplit_rcv_variant_stats_df_dict = {variant_name: [] for variant_name in rcv_variants.get_rcv_dict().keys()}
#     allsplit_rcv_group_stats_df_dict = {variant_group: [] for variant_group in
#                                         set(g.variant_group() for g in rcv_variants.get_rcv_dict().values())}

#     # init loggers
#     error_logger = init_error_logger(results_dir)

#     n_errors = 0
#     #########################
#     # LOOP TROUGH CONTESTS

#     for idx, contest in enumerate(contest_set):

#         # RUN OPERATIONS
#         pbar_desc = f'{idx+1} of {len(contest_set)} contests: '
#         pbar_desc += f'{contest["jurisdiction"]} {contest["date"]} {contest["office"]}'
#         if n_errors:
#             pbar_desc = f'[{n_errors} ERRORS SO FAR] ' + pbar_desc

#         step_obj = CrunchSteps(contest, output_config, converted_cvr_dir, results_dir, pbar_desc)
#         step_obj.update_error_log_writers([error_logger])
#         step_obj.run_steps()

#         crunch_returns = step_obj.return_results()
#         n_errors += crunch_returns['n_errors']

#         # STORE RESULTS
#         if output_config.get('candidate_details') and crunch_returns.get('candidate_details'):
#             candidate_details_dfs.append(crunch_returns['candidate_details'])

#         if output_config.get('per_rcv_type_stats') and crunch_returns.get('tabulation_stats_df'):
#             variant = crunch_returns['variant']
#             rcv_variant_stats_df_dict[variant].append(crunch_returns['tabulation_stats_df'])

#         if output_config.get('per_rcv_group_stats') and crunch_returns.get('contest_stats_df'):
#             variant_group = crunch_returns['variant_group']
#             rcv_group_stats_df_dict[variant_group].append(crunch_returns['contest_stats_df'])

#         if crunch_returns.get('split_stats'):

#             split_stats = crunch_returns['split_stats']

#             # SPLIT PATHS
#             split_path = results_dir / 'split_stats'
#             util.verifyDir(split_path)

#             split_contest_path = split_path / contest["uid"]
#             util.verifyDir(split_contest_path)

#             # SPLIT RESULTS CONTAINERS
#             split_rcv_variant_stats_df_dict = {variant_name: [] for variant_name in rcv_variants.get_rcv_dict().keys()}
#             split_rcv_group_stats_df_dict = {variant_group: [] for variant_group in
#                                              set(g.variant_group() for g in rcv_variants.get_rcv_dict().values())}

#             for split_stat in split_stats:

#                 if output_config.get('per_rcv_type_stats') and split_stat.get('tabulation_stats_df'):
#                     variant = split_stat['variant']
#                     split_rcv_variant_stats_df_dict[variant].append(split_stat['tabulation_stats_df'])
#                     allsplit_rcv_variant_stats_df_dict[variant].append(split_stat['tabulation_stats_df'])

#                 if output_config.get('per_rcv_group_stats') and split_stat.get('contest_stats_df'):
#                     variant_group = split_stat['variant_group']
#                     split_rcv_group_stats_df_dict[variant_group].append(split_stat['contest_stats_df'])
#                     allsplit_rcv_group_stats_df_dict[variant_group].append(split_stat['contest_stats_df'])

#             write_aggregated_stats(split_contest_path, output_config,
#                                    split_rcv_group_stats_df_dict, split_rcv_variant_stats_df_dict,
#                                    [], quiet=True)

#         # remove cvr from mem
#         if 'cvr' in contest:
#             del contest['cvr']

#     # close logs
#     error_logger.close()

#     # WRITE OUT AGGREGATED STATS FOR CONTESTS
#     write_aggregated_stats(results_dir,
#                            output_config,
#                            rcv_group_stats_df_dict,
#                            rcv_variant_stats_df_dict,
#                            candidate_details_dfs,
#                            quiet=False)

#     # WRITE OUT AGGREGATED STATS FOR ALL SPLITS
#     if os.path.isdir(results_dir / 'split_stats'):

#         allsplit_agg_path = results_dir / 'split_stats' / 'all_contest_splits'
#         util.verifyDir(allsplit_agg_path)

#         write_aggregated_stats(allsplit_agg_path,
#                                output_config,
#                                allsplit_rcv_group_stats_df_dict,
#                                allsplit_rcv_variant_stats_df_dict,
#                                [], quiet=True)

#     end_time = datetime.datetime.now()

#     write_input_dir(results_dir, output_config, start_time, end_time)

#     duration = end_time - start_time
#     print(f"runtime duration: {str(duration)}")
#     print("DONE!")




import csv
import pathlib
import json
import math
import os
import re
import collections
import decimal
import xmltodict

import pandas as pd

from rcv_cruncher.marks import BallotMarks

decimal.getcontext().prec = 30

global parser_dict


def add_parser(parser_dict) -> None:
    """Add custom parser functions to the module parser dictionary.

    :param parser_dict: A dictionary containing parser functions, with their names as keys.
    :type parser_dict: :data:`types.ParserDict`
    """
    parser_dict.update(parser_dict)


def get_parser_dict():
    """Returns the parser dictionary. Including both package parsers and
    custom parsers added with :func:`add_parser`.

    :return: A dictionary of parser functions. Keys are parser name strings.
    :rtype: ParserDict
    """
    return parser_dict


def cruncher_csv(cvr_path):
    """Reads ballot ranking information stored in csv format.
    One ballot per row, with ranking columns appearing in order and named with the word "rank"
    (e.x. "rank1", "rank2", etc)

    :param cvr_path: The path to the CVR file. If a file called "candidate_codes.csv" exists in the
    same directory, it will be read and columns named "code" and "candidate" will be used to replace
    candidate codes with candidate names in the CVR file during readin.
    :type cvr_path: :data:`types.Path`
    :raises RuntimeError: Error raised if not all parsed rank lists are the same length.
    :return: A dictionary of lists containing all columns in the CVR file.
        Rank columns are combined into per-ballot lists and stored with the key 'ranks'.
        A 'weight' key and list of 1's is added to the dictionary if no 'weight' column exists.
        All weights are of type :class:`decimal.Decimal`.
    :rtype: :data:`types.BallotDictOfLists`
    """

    cvr_path = pathlib.Path(cvr_path)
    df = pd.read_csv(cvr_path, encoding="utf8")

    # find rank columns
    rank_col = [col for col in df.columns if 'rank' in col.lower()]

    # ensure rank columns are strings
    df[rank_col] = df[rank_col].astype(str)

    # if candidate codes file exist, swap in names
    candidate_codes_fpath = cvr_path.parent / 'candidate_codes.csv'
    if os.path.isfile(candidate_codes_fpath):

        cand_codes = pd.read_csv(candidate_codes_fpath, encoding="utf8")
        cand_codes_dict = {str(code): cand for code, cand in zip(cand_codes['code'], cand_codes['candidate'])}
        replace_dict = {col: cand_codes_dict for col in rank_col}

        df = df.replace(replace_dict)

    # replace skipped ranks and overvotes with constants
    df = df.replace({col: {'under': BallotMarks.SKIPPED,
                           'skipped': BallotMarks.SKIPPED,
                           'undervote': BallotMarks.SKIPPED,
                           'over': BallotMarks.OVERVOTE,
                           'overvote': BallotMarks.OVERVOTE,
                           'UWI': BallotMarks.WRITEIN}
                    for col in rank_col})

    df = df.fillna(BallotMarks.SKIPPED)

    # pull out rank lists
    rank_col_list = [df[col].tolist() for col in rank_col]
    rank_lists = [list(rank_tuple) for rank_tuple in list(zip(*rank_col_list))]

    # double check all ballot ranks are equal length
    if not all([len(i) == len(rank_lists[0]) for i in rank_lists]):
        raise RuntimeError("not all rank lists are same length. debug")

    # assemble dict
    dct = {'ranks': rank_lists}

    # add in non-rank columns
    for col in df.columns:
        if col not in rank_col:
            dct[col] = df[col].tolist()

    # add weight if not present in csv
    if 'weight' not in dct:
        dct['weight'] = [decimal.Decimal('1') for b in dct['ranks']]
    else:
        dct['weight'] = [decimal.Decimal(str(w)) for w in dct['weight']]

    return dct


def dominion5_4(cvr_path, office):
    """Reads ballot data from Dominion V5.4 CVRs for a single contest.

    :param cvr_path: [description]
    :type cvr_path: :data:`types.Path`
    :param office: Names which contest's ballots should be read.
        Must match a contest name in ContestManifest.json.
    :type office: str
    :raises RuntimeError: If ballotIDs pulled from ImageMask field are not unique.
        Or if regex used to pull ballotID from ImageMask field malfunctions.
    :return:
    :rtype: :data:`types.BallotDictOfLists`
    """

    path = pathlib.Path(cvr_path)

    # load manifests, with ids as keys
    with open(path / 'ContestManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            if i['Description'] == office:
                current_contest_id = i['Id']
                current_contest_rank_limit = i['NumOfRanks']

    candidate_manifest = {}
    with open(path / 'CandidateManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            candidate_manifest[i['Id']] = i['Description']

    precinctPortion_manifest = {}
    with open(path / 'PrecinctPortionManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            precinctPortion_manifest[i['Id']] = {'Portion': i['Description'], 'PrecinctId': i['PrecinctId']}

    precinct_manifest = {}
    if os.path.isfile(path / 'PrecinctManifest.json'):
        with open(path / 'PrecinctManifest.json', encoding="utf8") as f:
            for i in json.load(f)['List']:
                precinct_manifest[i['Id']] = i['Description']

    ballotType_manifest = {}
    with open(path / 'BallotTypeManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            ballotType_manifest[i['Id']] = i['Description']

    countingGroup_manifest = {}
    with open(path / 'CountingGroupManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            countingGroup_manifest[i['Id']] = i['Description']

    ballotTypeContest_manifest = {}
    with open(path / 'BallotTypeContestManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:

            if i['ContestId'] not in ballotTypeContest_manifest.keys():
                ballotTypeContest_manifest[i['ContestId']] = []

            ballotTypeContest_manifest[i['ContestId']].append(i['BallotTypeId'])

    # read in ballots
    ballot_ranks = []
    ballot_IDs = []
    ballot_precinctPortions = []
    ballot_precincts = []
    ballot_types = []
    ballot_countingGroups = []
    with open(path / 'CvrExport.json', encoding="utf8") as f:
        for contests in json.load(f)['Sessions']:

            # ballotID
            ballotID_search = re.search('Images\\\\(.*)\*\.\*', contests['ImageMask'])
            if ballotID_search:
                ballotID = ballotID_search.group(1)
            else:
                raise RuntimeError('regex is not working correctly. debug')

            countingGroup = countingGroup_manifest[contests['CountingGroupId']]

            # for each session use original, or if isCurrent is False,
            # use modified
            if contests['Original']['IsCurrent']:
                current_contests = contests['Original']
            else:
                current_contests = contests['Modified']

            # precinctId for this ballot
            precinctPortion = precinctPortion_manifest[current_contests['PrecinctPortionId']]['Portion']
            precinctId = precinctPortion_manifest[current_contests['PrecinctPortionId']]['PrecinctId']

            precinct = None
            if precinct_manifest:
                precinct = precinct_manifest[precinctId]

            # ballotType for this ballot
            ballotType = ballotType_manifest[current_contests['BallotTypeId']]

            if len(current_contests['Cards']) > 1:
                print('"Cards" has length greater than 1, not prepared for this. debug')
                exit(1)

            ballot_contest_marks = None
            for ballot_contest in current_contests['Cards'][0]['Contests']:
                if ballot_contest['Id'] == current_contest_id:
                    ballot_contest_marks = ballot_contest['Marks']

            # skip ballot if didn't contain contest
            if ballot_contest_marks is None:
                continue

            # check for marks on each rank expected for this contest
            currentRank = 1
            current_ballot_ranks = []
            while currentRank <= current_contest_rank_limit:

                # find any marks that have the currentRank and aren't Ambiguous
                currentRank_marks = [i for i in ballot_contest_marks
                                     if i['Rank'] == currentRank and i['IsAmbiguous'] is False]

                if len(currentRank_marks) == 0:
                    currentCandidate = BallotMarks.SKIPPED
                elif len(currentRank_marks) > 1:
                    currentCandidate = BallotMarks.OVERVOTE
                else:
                    currentCandidate = candidate_manifest[currentRank_marks[0]['CandidateId']]

                current_ballot_ranks.append(currentCandidate)
                currentRank += 1

            ballot_ranks.append(current_ballot_ranks)
            ballot_precinctPortions.append(precinctPortion)
            ballot_precincts.append(precinct)
            ballot_IDs.append(ballotID)
            ballot_types.append(ballotType)
            ballot_countingGroups.append(countingGroup)

    ballot_dict = {'ranks': ballot_ranks,
                   'weight': [decimal.Decimal('1')] * len(ballot_ranks),
                   'ballotID': ballot_IDs,
                   'precinctPortion': ballot_precinctPortions,
                   'ballot_type': ballot_types,
                   'countingGroup': ballot_countingGroups}

    # make sure precinctManifest was part of CVR, otherwise exclude precinct column
    if len(ballot_precincts) != sum(i is None for i in ballot_precincts):
        ballot_dict['precinct'] = ballot_precincts

    # check ballotIDs are unique
    if len(set(ballot_dict['ballotID'])) != len(ballot_dict['ballotID']):
        raise RuntimeError("some non-unique ballot IDs")

    return ballot_dict


def dominion5_10(cvr_path, office):

    path = pathlib.Path(cvr_path)

    # load manifests, with ids as keys
    with open(path / 'ContestManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            if i['Description'] == office:
                current_contest_id = i['Id']
                current_contest_rank_limit = i['NumOfRanks']

    candidate_manifest = {}
    with open(path / 'CandidateManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            candidate_manifest[i['Id']] = i['Description']

    precinctPortion_manifest = {}
    with open(path / 'PrecinctPortionManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            precinctPortion_manifest[i['Id']] = {'Portion': i['Description'], 'PrecinctId': i['PrecinctId']}

    precinct_manifest = {}
    if os.path.isfile(path / 'PrecinctManifest.json'):
        with open(path / 'PrecinctManifest.json', encoding="utf8") as f:
            for i in json.load(f)['List']:
                precinct_manifest[i['Id']] = i['Description']

    district_manifest = {}
    with open(path / 'DistrictManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            district_manifest[i['Id']] = {'District': i['Description'], 'DistrictTypeId': i['DistrictTypeId']}

    districtType_manifest = {}
    with open(path / 'DistrictTypeManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            districtType_manifest[i['Id']] = i['Description']

    districtPrecinctPortion_manifest = {}
    with open(path / 'DistrictPrecinctPortionManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            districtPrecinctPortion_manifest[i['PrecinctPortionId']] = i['DistrictId']

    ballotType_manifest = {}
    with open(path / 'BallotTypeManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            ballotType_manifest[i['Id']] = i['Description']

    countingGroup_manifest = {}
    with open(path / 'CountingGroupManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            countingGroup_manifest[i['Id']] = i['Description']

    ballotTypeContest_manifest = {}
    with open(path / 'BallotTypeContestManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:

            if i['ContestId'] not in ballotTypeContest_manifest.keys():
                ballotTypeContest_manifest[i['ContestId']] = []

            ballotTypeContest_manifest[i['ContestId']].append(i['BallotTypeId'])

    tabulator_manifest = {}
    with open(path / 'TabulatorManifest.json', encoding="utf8") as f:
        for i in json.load(f)['List']:
            tabulator_manifest[i['Id']] = i['VotingLocationName']

    # read in ballots
    ballot_ranks = []
    ballot_IDs = []
    ballot_precinctPortions = []
    ballot_precincts = []
    ballot_types = []
    ballot_countingGroups = []
    ballot_votingLocation = []
    ballot_district = []
    ballot_districtType = []

    for cvr_export in path.glob("CvrExport*.json"):
        with open(cvr_export, encoding="utf8") as f:
            for contests in json.load(f)['Sessions']:

                # ballotID
                ballotID_search = re.search('Images\\\\(.*)\*\.\*', contests['ImageMask'])
                if ballotID_search:
                    ballotID = ballotID_search.group(1)
                else:
                    raise RuntimeError('regex is not working correctly. debug')

                countingGroup = countingGroup_manifest[contests['CountingGroupId']]

                # voting location for ballots
                ballotVotingLocation = tabulator_manifest[contests['TabulatorId']]

                # for each session use original, or if isCurrent is False,
                # use modified
                if contests['Original']['IsCurrent']:
                    current_contests = contests['Original']
                else:
                    current_contests = contests['Modified']

                # precinctId for this ballot
                precinctPortion = precinctPortion_manifest[current_contests['PrecinctPortionId']]['Portion']
                precinctId = precinctPortion_manifest[current_contests['PrecinctPortionId']]['PrecinctId']

                precinct = None
                if precinct_manifest:
                    precinct = precinct_manifest[precinctId]

                # ballotType for this ballot
                ballotType = ballotType_manifest[current_contests['BallotTypeId']]

                # district for ballot
                ballotDistrictId = districtPrecinctPortion_manifest[current_contests['PrecinctPortionId']]
                ballotDistrict = district_manifest[ballotDistrictId]['District']
                ballotDistrictType = districtType_manifest[district_manifest[ballotDistrictId]['DistrictTypeId']]

                ballot_contest_marks = None
                for cards in current_contests['Cards']:
                    for ballot_contest in cards['Contests']:
                        if ballot_contest['Id'] == current_contest_id:
                            if ballot_contest_marks is not None:
                                raise (RuntimeError(
                                    "Contest Id appears twice across a single set of cards. Not expected."))
                            ballot_contest_marks = ballot_contest['Marks']

                # skip ballot if didn't contain contest
                if ballot_contest_marks is None:
                    continue

                # check for marks on each rank expected for this contest
                currentRank = 1
                current_ballot_ranks = []
                while currentRank <= current_contest_rank_limit:

                    # find any marks that have the currentRank and aren't Ambiguous
                    currentRank_marks = [i for i in ballot_contest_marks
                                         if i['Rank'] == currentRank and i['IsAmbiguous'] is False]

                    currentCandidate = '**error**'

                    if len(currentRank_marks) == 0:
                        currentCandidate = BallotMarks.SKIPPED
                    elif len(currentRank_marks) > 1:
                        currentCandidate = BallotMarks.OVERVOTE
                    else:
                        currentCandidate = candidate_manifest[currentRank_marks[0]['CandidateId']]

                    if currentCandidate == '**error**':
                        raise RuntimeError('error in filtering marks. debug')

                    current_ballot_ranks.append(currentCandidate)
                    currentRank += 1

                ballot_ranks.append(current_ballot_ranks)
                ballot_precinctPortions.append(precinctPortion)
                ballot_precincts.append(precinct)
                ballot_IDs.append(ballotID)
                ballot_types.append(ballotType)
                ballot_countingGroups.append(countingGroup)
                ballot_votingLocation.append(ballotVotingLocation)
                ballot_district.append(ballotDistrict)
                ballot_districtType.append(ballotDistrictType)

    ballot_dict = {'ranks': ballot_ranks,
                   'weight': [decimal.Decimal('1')] * len(ballot_ranks),
                   'ballotID': ballot_IDs,
                   'precinct': ballot_precincts,
                   'precinctPortion': ballot_precinctPortions,
                   'ballot_type': ballot_types,
                   'countingGroup': ballot_countingGroups,
                   'votingLocation': ballot_votingLocation,
                   'district': ballot_district,
                   'districtType': ballot_districtType}

    # make sure precinctManifest was part of CVR, otherwise exclude precinct column
    if len(ballot_precincts) != sum(i is None for i in ballot_precincts):
        ballot_dict['precinct'] = ballot_precincts

    # check ballotIDs are unique
    if len(set(ballot_dict['ballotID'])) != len(ballot_dict['ballotID']):
        raise RuntimeError("some non-unique ballot IDs")

    return ballot_dict


def choice_pro_plus(cvr_path):

    cvr_path = pathlib.Path(cvr_path)

    # read chp file
    chp_glob = [f for f in cvr_path.glob('*.chp')]
    if len(chp_glob) > 1:
        raise RuntimeError(f'more than one .chp file found in directory {str(cvr_path)}')

    candidate_map = {}
    prm_files = []
    with open(chp_glob[0], encoding='utf8') as f:
        for i in f:

            split = i.strip('\n').split()

            if len(split) >= 3 and split[0] == '.CANDIDATE':
                candidate_map[split[1].strip(',')] = i.split('"')[1].split('"')[0]

            if len(split) == 2 and split[0] == '.INCLUDE':
                prm_files.append(cvr_path / split[1])

    # read prm files
    if not prm_files:
        raise RuntimeError(f'no .prm files listed in .chp file {chp_glob[0]}')

    ballots = []
    for prm_file in prm_files:
        with open(prm_file, 'r', encoding='utf8') as f:
            for i in f:
                if any(map(str.isalnum, i)) and i.strip()[0] != '#':
                    b = []
                    s = i.split()
                    choices = [] if len(s) == 1 else s[1].split(',')
                    for choice in filter(None, choices):
                        can, rank = choice.split(']')[0].split('[')
                        b.extend([BallotMarks.SKIPPED] * (int(rank) - len(b) - 1))
                        b.append(BallotMarks.OVERVOTE if '=' in choice else candidate_map[can])
                    ballots.append(b)

    # add in tail skipped ranks
    maxlen = max(map(len, ballots))
    for b in ballots:
        b.extend([BallotMarks.SKIPPED] * (maxlen - len(b)))

    return {'ranks': ballots, 'weight': [decimal.Decimal('1')] * len(ballots)}


def burlington2006(cvr_path):
    """Function to parse file format used in 2006 Burlington Mayoral Election.

    Args:
        ctx (dictionary): A dictionary containing key "path" which points to the cvr file
        containing ranking information of each ballot. If a file called "candidate_codes.csv"
        exists in the same directory, it will be read and values from the "code" column
        that are found in the cvr file ranks will be replaced values in the "candidate" columns.

    Returns:
        dictionary: Dictionary with a single key, "ranks", containing parsed ballots as a list of lists.
        Overvotes and skipped rankings are represented by constants.
    """
    path = pathlib.Path(cvr_path)

    # read in lines
    ballots = []
    with open(path, "r", encoding='utf8') as f:
        for line in f:
            ballots.append([BallotMarks.OVERVOTE if '=' in i else i for i in line.split()[3:]])

    # fill in skipped ranks with constant
    maxlen = max(map(len, ballots))
    for b in ballots:
        b.extend([BallotMarks.SKIPPED] * (maxlen - len(b)))

    # read candidate codes
    candidate_codes_fname = path.parent / "candidate_codes.csv"
    if os.path.exists(candidate_codes_fname):

        cand_codes = pd.read_csv(candidate_codes_fname, encoding="utf8")
        cand_codes_dict = {str(code): cand for code, cand in zip(cand_codes['code'], cand_codes['candidate'])}

        # replace candidate codes with candidate names
        new_ballots = []
        for b in ballots:
            new_ballots.append([cand_codes_dict[cand] if cand in cand_codes_dict else cand for cand in b])

    return {'ranks': new_ballots}


def optech1(cvr_path, office):

    cvr_path = pathlib.Path(cvr_path)

    # FIND THE FILES
    ballot_image_files = [f for f in cvr_path.glob('*allot*.txt')]
    master_lookup_files = [f for f in cvr_path.glob('*aster*.txt')]

    ballot_image_path = None
    if not ballot_image_files:
        raise RuntimeError(f'parser error - no ballot image file found in {cvr_path}')
    elif len(ballot_image_files) > 1:
        raise RuntimeError(f'too many ballot image files in directory {cvr_path}. Should only be one.')
    else:
        ballot_image_path = ballot_image_files[0]

    master_lookup_path = None
    if not master_lookup_files:
        raise RuntimeError(f'parser error - no master lookup file found in {cvr_path}')
    elif len(master_lookup_files) > 1:
        raise RuntimeError(f'too many master lookup files in directory {cvr_path}. Should only be one.')
    else:
        master_lookup_path = master_lookup_files[0]

    # READ MASTER LOOKUP
    master_lookup = collections.defaultdict(dict)
    candidate_contest_map = {}
    with open(master_lookup_path, encoding='utf8') as f:
        for i in f:
            mapping = i[:10].strip()
            key = i[10:17].strip()
            value = i[17:67].strip()

            master_lookup[mapping][key] = value

            if mapping == "Candidate":
                candidate_contest_id = i[74:81].strip()
                candidate_contest_map.update({key: candidate_contest_id})

    # find contest id
    contest_reverse_map = {v: k for k, v in master_lookup['Contest'].items()}
    if office not in contest_reverse_map:
        raise RuntimeError(f'contest set office value ({office}) not present in master lookup file {master_lookup_path}')
    contest_id = contest_reverse_map[office]

    # remove candidates from master lookup if they are from another contest
    master_lookup['Candidate'] = {candidate_id: candidate_val for candidate_id, candidate_val
                                  in master_lookup['Candidate'].items()
                                  if candidate_contest_map[candidate_id] == contest_id}

    # tally types are stored in master lookup with 7 chars but only recorded in ballot image with 3
    # trim off the first 4 char from the master lookup strings
    tally_type_map = {k[4:]: v for k, v in master_lookup['Tally Type'].items()}

    # separate out other maps
    precinct_map = master_lookup['Precinct']
    name_map = {k: {'WRITEIN': BallotMarks.WRITEIN}.get(v.upper().replace('-', ''), v)
                for k, v in master_lookup['Candidate'].items()}

    # READ BALLOT FILE
    voter_info_collected = collections.defaultdict(list)
    max_rank_num = 0
    with open(ballot_image_path, "r", encoding='utf8') as f:

        for line in f:

            line_contest_id = line[:7].strip()
            line_voter_id = line[7:16].strip()
            line_tally_type = line[23:26].strip()
            line_precinct_id = line[26:33].strip()
            line_candidate_id = line[36:43].strip()
            line_rank = line[33:36].strip()
            line_skipped = line[43].strip()
            line_overvote = line[44].strip()

            # skip line if not for contest
            if line_contest_id != contest_id:
                continue

            max_rank_num = int(line_rank) if int(line_rank) > max_rank_num else max_rank_num

            voter_info_collected[line_voter_id].append({
                'voter_id': line_voter_id,
                'tally_type': tally_type_map[line_tally_type],
                'precinct': precinct_map[line_precinct_id],
                # 0 candidate id plus a skipped or overvote mark, indicate skip or overvote
                'candidate': name_map[line_candidate_id] if int(line_candidate_id) else 0,
                'rank': int(line_rank),
                'skipped': int(line_skipped),
                'overvote': int(line_overvote)
                })

            # debug check to see if ballot marks with valid candidate name stored sometimes also get paired,
            # with overvote or skipped marks?
            if int(line_candidate_id) and (int(line_skipped) or int(line_overvote)):
                raise RuntimeError('both a skip and overvote mark for this rank. unexpected')

    # for each voter, assemble the ballot
    dct = {
        'ranks': [],
        'precinct': [],
        'tally_type': [],
        'ballotID': []
    }
    for voter_id in voter_info_collected:

        voter_ballot_info = voter_info_collected[voter_id]

        # debug checks
        voter_tally_type_set = set([b['tally_type'] for b in voter_ballot_info])
        voter_precinct_set = set([b['precinct'] for b in voter_ballot_info])

        if len(voter_tally_type_set) > 1:
            raise RuntimeError("Marks for this voter contain multiple tally type values. Unexpected.")

        if len(voter_precinct_set) > 1:
            raise RuntimeError("Marks for this voter contain multiple precinct values. Unexpected.")

        voter_tally_type = list(voter_tally_type_set)[0]
        voter_precinct = list(voter_precinct_set)[0]

        voter_ranks = [None] * max_rank_num
        for rank in range(1, max_rank_num + 1):

            rank_info_filter = [b for b in voter_ballot_info if b['rank'] == rank]

            if len(rank_info_filter) > 1:
                raise RuntimeError('unexpected')
                # an overvote marker is stored in the file, but do overvoted ranks still get all their marks stored ?
            elif len(rank_info_filter) == 0:
                raise RuntimeError('unexpected')
                # this would be unexpected since there is a skipped rank mark in the file
            else:
                rank_info = rank_info_filter[0]

            rank_candidate = rank_info['candidate']

            if rank_candidate == 0 and rank_info['skipped'] and rank_info['overvote']:
                raise RuntimeError('this shouldnt be reached')
            elif rank_candidate == 0 and rank_info['skipped']:
                voter_ranks[rank-1] = BallotMarks.SKIPPED
            elif rank_candidate == 0 and rank_info['overvote']:
                voter_ranks[rank-1] = BallotMarks.OVERVOTE
            else:
                voter_ranks[rank-1] = rank_candidate

        if any(r is None for r in voter_ranks):
            raise RuntimeError('not all ranks for this voter had data stored in the file. unexpected.')

        dct['ranks'].append(voter_ranks)
        dct['ballotID'].append(voter_id)
        dct['tally_type'].append(voter_tally_type)
        dct['precinct'].append(voter_precinct)

    # add weights
    dct.update({'weight': [decimal.Decimal('1')] * len(dct['ranks'])})
    return dct


def optech2(cvr_path):

    cvr_path = pathlib.Path(cvr_path)

    ballot_glob = [f for f in cvr_path.glob('*allot*.txt')]
    cntl_glob = [f for f in cvr_path.glob('*ntl*.txt')]

    if len(ballot_glob) > 1:
        raise RuntimeError(f'more than 1 file with pattern matching *allot*.txt found in directory {str(cvr_path)}')

    if len(cntl_glob) > 1:
        raise RuntimeError(f'more than 1 file with pattern matching *ntl*.txt found in directory {str(cvr_path)}')

    # read in canadidate codes and names
    candidate_map = {}
    with open(cntl_glob[0], encoding='utf8') as f:
        for i in f:
            line = [j.strip() for j in i.split(':')]
            if line and line[0] == 'Candidate':
                candidate_map[line[1]] = line[2]

    candidate_map['--'] = BallotMarks.SKIPPED
    candidate_map['++'] = BallotMarks.OVERVOTE

    # read ballots
    ballots = []
    with open(ballot_glob[0], "r", encoding='utf8') as f:
        line = f.readline()
        while line:
            ballots.append([candidate_map[i] for i in line.split()[-1].split('>')])
            line = f.readline()

    return ballots


def minneapolis2009(cvr_path, office):

    cvr_path = pathlib.Path(cvr_path)

    # read map file
    map_file = cvr_path.parent / 'convert.csv'

    choice_map = {}
    default = None
    with open(map_file, encoding='utf8') as f:
        for i in f:
            split = i.strip().split('\t')
            if len(split) >= 3 and split[0] == office:
                choice_map[split[2]] = split[1]

    if choice_map == {}:
        raise RuntimeError('No candidates found. Ensure "office" field in contest_set matches CVR.')

    choice_map['XXX'] = BallotMarks.SKIPPED
    default = BallotMarks.WRITEIN

    # read ballots
    precincts = []
    ballots = []
    with open(cvr_path, "r", encoding='utf8') as f:
        f.readline()
        for line in csv.reader(f):
            choices = [choice_map.get(i.strip(), i if default is None else default)
                       for i in line[1:-1]]
            if choices != ['', '', '']:
                ballots.extend([choices] * int(float(line[-1])))
                for p in range(int(float(line[-1]))):
                    precincts.append(line[0])

    bs = {'ranks': ballots,
          'weight': [decimal.Decimal('1')] * len(ballots),
          'precinct': precincts}

    return bs


# def santafe(column_id, contest_id, ctx):
#     path = ctx['cvr_path']
#     candidate_map = {}
#     with open(ctx['cvr_path'].replace('CvrExport', 'CandidateManifest'), encoding='utf8') as f:
#         for i in f:
#             row = i.split(',')
#             if row:
#                 candidate_map[row[1]] = row[0]
#     ballots = []
#     ballot_length = 0
#     with open(path, "r", encoding='utf8') as f:
#         reader = csv.reader(f)
#         header = next(reader)
#         s = 'Original/Cards/0/Contests/{}/Marks/{}/{}'
#         rinds = []
#         cinds = []
#         for i in range(len(header)):
#             try:
#                 rinds.append(header.index(s.format(column_id, i, 'Rank')))
#             except ValueError:
#                 break
#             cinds.append(header.index(s.format(column_id, i, 'CandidateId')))
#         col = header.index('Original/Cards/0/Contests/{}/Id'.format(column_id))
#         for line in reader:
#             if line[col] == str(contest_id):
#                 choices = []
#                 ranks = [int(line[i]) for i in rinds if line[i] != '']
#                 ballot_length = max(ranks + [ballot_length])
#                 candidates = iter(cinds)
#                 for i in range(len(rinds)):
#                     c = ranks.count(i+1)
#                     if c == 0:
#                         choices.append(BallotMarks.SKIPPED)
#                     elif c == 1:
#                         next_candidate = line[next(candidates)]
#                         choices.append(candidate_map[next_candidate])
#                     else:
#                         choices.append(BallotMarks.OVERVOTE)
#                 ballots.append(choices)
#     return [b[:ballot_length] for b in ballots]


# def santafe_id(column_id, contest_id, ctx):
#     path = ctx['cvr_path']
#     ballots = []
#     ballot_length = 0
#     with open(path, "r", encoding='utf8') as f:
#         reader = csv.reader(f)
#         header = next(reader)
#         s = 'Original/Cards/0/Contests/{}/Marks/{}/{}'
#         rinds = []
#         cinds = []
#         for i in range(len(header)):
#             try:
#                 rinds.append(header.index(s.format(column_id, i, 'Rank')))
#             except ValueError:
#                 break
#             cinds.append(header.index(s.format(column_id, i, 'CandidateId')))
#         col = header.index('Original/Cards/0/Contests/{}/Id'.format(column_id))
#         for i, line in enumerate(reader):
#             if line[col] == str(contest_id):
#                 choices = collections.UserList([])
#                 choices.voter_id = i
#                 ranks = [int(line[i]) for i in rinds if line[i] != '']
#                 ballot_length = max(ranks + [ballot_length])
#                 candidates = iter(cinds)
#                 for i in range(len(rinds)):
#                     c = ranks.count(i+1)
#                     if c == 0:
#                         choices.append(BallotMarks.SKIPPED)
#                     elif c == 1:
#                         choices.append(line[next(candidates)])
#                     else:
#                         choices.append(BallotMarks.OVERVOTE)
#                 ballots.append(choices)
#     for b in ballots:
#         b.data = b.data[:ballot_length]
#     return ballots


def dominion5_2(cvr_path, office):

    path = pathlib.Path(cvr_path)

    with open(path / 'ContestManifest.json', encoding='utf8') as f:
        for i in json.load(f)['List']:
            if i['Description'] == office.upper():
                contest_id = i['Id']
                ranks = i['NumOfRanks']
                if ranks == 0:
                    ranks = 1

    candidates = {}
    with open(path / 'CandidateManifest.json', encoding='utf8') as f:
        for i in json.load(f)['List']:
            if i['ContestId'] == contest_id:
                candidates[i['Id']] = i['Description']

    precincts = {}
    with open(path / 'PrecinctPortionManifest.json', encoding='utf8') as f:
        for i in json.load(f)['List']:
            precincts[i['Id']] = i['Description'].split()[1]

    ballotType_manifest = {}
    with open(path / 'BallotTypeManifest.json', encoding='utf8') as f:
        for i in json.load(f)['List']:
            ballotType_manifest[i['Id']] = i['Description']

    countingGroup_manifest = {}
    with open(path / 'CountingGroupManifest.json', encoding='utf8') as f:
        for i in json.load(f)['List']:
            countingGroup_manifest[i['Id']] = i['Description']

    ballots = {'ranks': [], 'ballotID': [], 'precinct': [], 'ballotType': [], 'countingGroup': [], 'weight': []}
    with open(path / 'CvrExport.json', encoding='utf8') as f:

        for contests in json.load(f)['Sessions']:

            # ballotID
            ballotID_search = re.search('Images\\\\(.*)\*\.\*', contests['ImageMask'])
            if ballotID_search:
                ballotID = ballotID_search.group(1)
            else:
                raise RuntimeError('regex is not working correctly. debug')

            countingGroup = countingGroup_manifest[contests['CountingGroupId']]

            if contests['Original']['IsCurrent']:
                current_contests = contests['Original']
            else:
                current_contests = contests['Modified']

            precinct = precincts[current_contests['PrecinctPortionId']]
            ballotType = ballotType_manifest[current_contests['BallotTypeId']]

            for contest in current_contests['Contests']:

                # confirm correct contest
                if contest['Id'] == contest_id:

                    # make empty ballot
                    ballot = [BallotMarks.SKIPPED] * ranks

                    # look through marks
                    for mark in contest['Marks']:
                        candidate = candidates[mark['CandidateId']]
                        if candidate == 'Write-in':
                            candidate = BallotMarks.WRITEIN
                        rank = mark['Rank']-1
                        if mark['IsAmbiguous']:
                            pass
                        elif ballot[rank] == BallotMarks.OVERVOTE:
                            pass
                        elif ballot[rank] == BallotMarks.SKIPPED:
                            ballot[rank] = candidate
                        elif ballot[rank] != candidate:
                            ballot[rank] = BallotMarks.OVERVOTE

                    ballots['countingGroup'].append(countingGroup)
                    ballots['ballotType'].append(ballotType)
                    ballots['precinct'].append(precinct)
                    ballots['ranks'].append(ballot)
                    ballots['ballotID'].append(ballotID)

    ballots['weight'] = [decimal.Decimal('1')] * len(ballots['ranks'])

    # check ballotIDs are unique
    if len(set(ballots['ballotID'])) != len(ballots['ballotID']):
        print("some non-unique ballot IDs")
        exit(1)

    return ballots


def unisyn(cvr_path):
    """
    This parser was developed for the unisyn 2020 Hawaii Dem Primary CVR which only contained the
    ranked choice votes for a single election. Unisyn uses the common data format in xml, however the
    parser currently is not a complete common data format parser.

    For more information on common data format, see:
    https://pages.nist.gov/CastVoteRecords/
    https://github.com/hiltonroscoe/cdfprototype
    """

    glob_str = pathlib.Path(cvr_path).glob('/*.xml')

    contestIDdicts = {}
    for f in glob_str:

        with open(f) as fd:
            xml_dict = xmltodict.parse(fd.read())

        # get candidates
        candidatesIDs = {}
        for cand_dict in xml_dict['CastVoteRecordReport']['Election']['Candidate']:
            candidatesIDs[cand_dict['@ObjectId']] = cand_dict['Name']

        # loop through CVR snapshots
        for cvr_dict in xml_dict['CastVoteRecordReport']['CVR']:

            cvr_contest = cvr_dict['CVRSnapshot']['CVRContest']

            if cvr_contest['ContestId'] not in contestIDdicts:
                contestIDdicts.update({cvr_contest['ContestId']: []})
            contestIDdicts[cvr_contest['ContestId']].append(cvr_contest['CVRContestSelection'])

    # check that all rank lists are equal
    first_rank = list(contestIDdicts.keys())[0]
    rank_length_equal = [len(contestIDdicts[k]) == len(contestIDdicts[first_rank]) for k in contestIDdicts]
    if not all(rank_length_equal):
        print("not all rank lists are equal.")
        raise RuntimeError

    # combine ranks into lists
    ballot_lists = []
    for idx in range(len(contestIDdicts[first_rank])):

        idx_ranks = [int(contestIDdicts[rank_key][idx]['Rank']) for rank_key in contestIDdicts]

        idx_candidates = []
        for rank_key in contestIDdicts:
            contest_dict = contestIDdicts[rank_key][idx]
            if 'SelectionPosition' in contest_dict:
                if isinstance(contest_dict['SelectionPosition'], list):
                    idx_candidates.append(BallotMarks.OVERVOTE)
                else:
                    idx_candidates.append(candidatesIDs[contest_dict['SelectionPosition']['Position']])
            elif contest_dict['TotalNumberVotes'] == '0':
                idx_candidates.append(BallotMarks.SKIPPED)

        ordered_ranks = sorted(zip(idx_candidates, idx_ranks), key=lambda x: x[1])
        ballot_lists.append([t[0] for t in ordered_ranks])

    # assemble dict
    dct = {'ranks': ballot_lists}
    dct['weight'] = [decimal.Decimal('1')] * len(dct['ranks'])

    return dct


def surveyUSA(cvr_path):
    """
    Survey USA files usually include all respondents and should be pre-filtered for any columns
    prior to cruncher use (such as filtering likely democratic voters). Rank columns and ballotID columns
    should also be renamed prior to parsing.

    Rank columns can contain candidate codes or NaN. NaN is treated as a skipped rank.

    Required files:
    cvr.csv - contains ballots, ranks, weights
    candidate_codes.csv - contains two columns ("code" and "candidate") that map cvr code numbers to candidate names.
    """

    path = pathlib.Path(cvr_path)

    csv_df = pd.read_csv(path / 'cvr.csv')
    candidate_codes_df = pd.read_csv(path / 'candidate_codes.csv')

    # candidate code dict
    candidate_map = {row['code']: row['candidate'] for index, row in candidate_codes_df.iterrows()}

    # find rank columns
    rank_columns = [col for col in csv_df.columns if 'rank' in col.lower()]

    ballots = []
    for index, row in csv_df.iterrows():

        b_ranks = [BallotMarks.SKIPPED] * len(rank_columns)

        saw_undecided = False
        since_undecided = []

        for idx, rank in enumerate(rank_columns):

            # nan marks end of ranks
            if math.isnan(row[rank]):
                if since_undecided:
                    print('some candidates appeared after an undecided vote! debug')
                    raise RuntimeError
                break

            candidate = candidate_map[row[rank]]

            if saw_undecided:
                since_undecided.append(candidate)

            if candidate == 'Undecided':
                saw_undecided = True

            if candidate != 'Undecided':
                b_ranks[idx] = candidate

        ballots.append(b_ranks)

    ballot_dict = {'ranks': ballots, 'weight': csv_df['weight'], 'ballotID': csv_df['ballotID']}
    return ballot_dict


parser_dict = {
    "burlington2006": burlington2006,
    "cruncher_csv": cruncher_csv,
    "dominion5_2": dominion5_2,
    "dominion5_4": dominion5_4,
    "dominion5_10": dominion5_10,
    "optech": optech1,
    "unisyn": unisyn,
    "surveyUSA": surveyUSA,
    "minneapolis2009": minneapolis2009
    # "santafe": santafe, still need to figure out this parser
    # "santafe_id": santafe_id,
}
